/**
 *   @file  qmss_drv.c
 *
 *   @brief   
 *      This is the Queue Manager Low Level Driver file.
 *
 *  \par
 *  ============================================================================
 *  @n   (C) Copyright 2009-2013, Texas Instruments, Inc.
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  \par
*/

/* QMSS Types includes */
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>

/* QMSS LLD includes */
#include <ti/drv/qmss/qmss_drv.h>
#include <ti/drv/qmss/include/qmss_pvt.h>
#include <ti/drv/qmss/include/qmss_intd.h>

/* QMSS OSAL layer */
#include <qmss_osal.h>

/* RM includes */
#include <ti/drv/rm/rm_services.h>

/**********************************************************************
 ************************** Globals ***********************************
 **********************************************************************/

/* Maintain status of queues */
#ifdef _TMS320C6X
#pragma DATA_SECTION (qmssQueueFree, ".qmss");
#pragma DATA_ALIGN (qmssQueueFree, QMSS_MAX_CACHE_ALIGN)
#endif
static uint8_t              qmssQueueFree[QMSS_MAX_QUEUES];
/* It is assumed QMSS_MAX_QUEUES is divisible by 128, so no padding is needed */

#ifdef _TMS320C6X
#pragma DATA_SECTION (qmssGroupControl, ".qmss");
#pragma DATA_ALIGN (qmssGroupControl, QMSS_MAX_CACHE_ALIGN)
#endif
struct {
    /** Used to do load balanced allocation */
    int32_t groupFreeQueues[QMSS_MAX_QMGR_GROUPS];
    /** Used to perform round robin allocation */
    uint32_t alternate;
    /** Padding required to make sure linker doesn't put something else
     * on the same cache line.
     */
    uint8_t pad[QMSS_MAX_CACHE_ALIGN - sizeof(uint32_t) * (QMSS_MAX_QMGR_GROUPS + 1)];
} qmssGroupControl;

/* QMSS Global object */
#ifdef _TMS320C6X
#pragma DATA_SECTION (qmssGObjIsValid, ".qmss");
#pragma DATA_ALIGN (qmssGObjIsValid, QMSS_MAX_CACHE_ALIGN)
#pragma DATA_SECTION (qmssGObj, ".qmss");
#pragma DATA_ALIGN (qmssGObj, QMSS_MAX_CACHE_ALIGN)
/* force these two to a local memory when default is shared */
#pragma DATA_SECTION (qmssLObj, ".far:local");
#pragma DATA_SECTION (qmssLObjIsValid, ".far:local");
#endif
Qmss_GlobalObj              qmssGObj;

/* QMSS Local object */
Qmss_LocalObj               qmssLObj;

/* QMSS Global Object is valid flag */
uint8_t                     qmssGObjIsValid = 0;

/* QMSS Local Object is valid flag */
uint8_t _C6000_FAR_ATTR_    qmssLObjIsValid = 0;

/** @brief Global Variable which describes the QMSS LLD Version Information */
const char   qmssLldVersionStr[] = QMSS_LLD_VERSION_STR ":" __DATE__  ":" __TIME__;

/**********************************************************************
 ************************** Functions *********************************
 **********************************************************************/

static Qmss_QueueHnd Qmss_internalQueueOpen (int qGroupIsSpecified, uint32_t qGroup, Qmss_QueueType queType, int queNum, uint8_t *isAllocated);
static Qmss_Result Qmss_internaldownloadFirmware (Qmss_PdspId pdspId, void *image, uint32_t size);
Qmss_Result Qmss_internalQueueClose (Qmss_QueueHnd hnd);

/** @addtogroup QMSS_LLD_FUNCTION
@{ 
*/

/**
 *  @b Description
 *  @n  
 *      Internal function get the queue type of a specified queue
 */
static int32_t Qmss_internalFindQueueType (uint32_t queNum)
{
    int32_t retVal = -1;
    int32_t qType; 
    int32_t queTypeStart, queTypeEnd;

    for (qType = Qmss_QueueType_LOW_PRIORITY_QUEUE; qType <= Qmss_QueueType_GENERAL_PURPOSE_QUEUE; qType++)
    {
        /* Ignore queue types that don't exist for a queue group.  These queue types will have a startIndex
         * and maxNum of zero */
        if ((qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(queNum)][qType].startIndex != 0) ||
            (qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(queNum)][qType].maxNum != 0))
        {
            queTypeStart = qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(queNum)][qType].startIndex;
            queTypeEnd = qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(queNum)][qType].startIndex + 
                         qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(queNum)][qType].maxNum - 1;

            if ((queNum >= queTypeStart) && (queNum <= queTypeEnd))
            {  
                retVal = qType;
                break;
            }
        }
    }
    return (retVal);
}

/**
 *  @b Description
 *  @n  
 *      Internal function to increment or decrement the number of queues open
 *      in each group.
 */
static void Qmss_internalUpdateGroupControl (uint32_t qGroup, int sign)
{
    /* Invalidate the qmssGroupControl structure */
    Qmss_osalBeginMemAccess ((void *) &qmssGroupControl, sizeof(qmssGroupControl));

    /* One fewer free queues in group */
    qmssGroupControl.groupFreeQueues[qGroup] += sign;

    /* Writeback qmssGroupControl structure */
    Qmss_osalEndMemAccess ((void *) &qmssGroupControl, sizeof(qmssGroupControl));
}


/**
 *  @b Description
 *  @n  
 *      Internal function to program linking RAM configuration for each
 *      queue manager group
 */

static void Qmss_internalInsertLinkingRAM (
    uint32_t         group, 
    Qmss_LinkingRAM *linkingRAM,
    uint32_t        *actSize)
{
    if (linkingRAM->linkingRAM0Base == 0)
    {
        qmssLObj.p.groupRegs[group].qmConfigReg->LINKING_RAM_REGION_0_BASE_ADDRESS_REG = 
           (uint32_t)qmssLObj.p.regs.qmLinkingRAMReg - (uint32_t)qmssLObj.p.regs.qmBaseAddr;
    }
    else
    {
        qmssLObj.p.groupRegs[group].qmConfigReg->LINKING_RAM_REGION_0_BASE_ADDRESS_REG = 
            (uint32_t)Qmss_osalVirtToPhy((void *)linkingRAM->linkingRAM0Base);
    }

    if (linkingRAM->linkingRAM0Size == 0)
    {
        *actSize = QMSS_LINKING_RAM_REGION_0_DEFAULT_SIZE;
        qmssLObj.p.groupRegs[group].qmConfigReg->LINKING_RAM_REGION_0_SIZE_REG = QMSS_LINKING_RAM_REGION_0_DEFAULT_SIZE;
    }
    else
        qmssLObj.p.groupRegs[group].qmConfigReg->LINKING_RAM_REGION_0_SIZE_REG = linkingRAM->linkingRAM0Size;

    qmssLObj.p.groupRegs[group].qmConfigReg->LINKING_RAM_REGION_1_BASE_ADDRESS_REG = 
        (uint32_t)Qmss_osalVirtToPhy((void *)linkingRAM->linkingRAM1Base);
}

/**
 *  @b Description
 *  @n  
 *      Internal function to validate linking RAM configuration for each
 *      queue manager group
 *
 *  @retval
 *      Success -   QMSS_SOK
 *  @retval
 *      Failure -   QMSS_INVALID_PARAM
 */

static Qmss_Result Qmss_internalCheckLinkingRAM (Qmss_LinkingRAM *linkingRAM)
{
    /* Check if LinkingRAM0 can hold all the descriptors if LinkingRAM1 is NULL */
    if (linkingRAM->linkingRAM1Base == 0) 
    {
        if (linkingRAM->maxDescNum > (QMSS_LINKING_RAM_REGION_0_DEFAULT_SIZE + 1)) 
        {
            return QMSS_INVALID_PARAM;
        }
    }
    return QMSS_SOK;
}

/*
 *  @b Description
 *  @n  
 *      This function sends a service request to RM
 *
 *  @param[in]  type
 *      Service request type
 *  @param[in]  resName
 *      Pointer to the RM resource name
 *  @param[out]  resNum
 *      Pointer to the resource value to check.  If the input value is
 *      unspecified the resource value returned by RM will be returned 
 *      through this parameter.
 *  @param[in]  resLen
 *      The resource length (starting from resNum) to check.
 *  @param[out] isAllocated
 *      Returns the reference count for the resource
 *
 *  @retval
 *      1 - RM approved request.
 *      0 - RM denied request or experienced an error
 */
int Qmss_rmService (Rm_ServiceHandle *rmService, Rm_ServiceType type, const char *resName, 
                    int32_t *resNum, uint32_t resLen, int32_t resAlign, uint8_t *isAllocated)
{
    Rm_ServiceReqInfo   rmServiceReq;
    Rm_ServiceRespInfo  rmServiceResp;
    int                 retVal = 0;

    memset((void *)&rmServiceReq, 0, sizeof(rmServiceReq));
    memset((void *)&rmServiceResp, 0, sizeof(rmServiceResp));
    
    rmServiceReq.type = type;
    rmServiceReq.resourceName = resName;
    if (*resNum >= 0)
    {
        rmServiceReq.resourceBase = *resNum;
    }
    else
    {
        rmServiceReq.resourceBase = RM_RESOURCE_BASE_UNSPECIFIED;
    }
    rmServiceReq.resourceLength = resLen;
    rmServiceReq.resourceAlignment = resAlign;
    /* RM will block until resource is returned since callback is NULL */
    rmServiceReq.callback.serviceCallback = NULL;
    rmService->Rm_serviceHandler(rmService->rmHandle, &rmServiceReq, &rmServiceResp);
    if ((rmServiceResp.serviceState == RM_SERVICE_APPROVED) ||
        (rmServiceResp.serviceState == RM_SERVICE_APPROVED_STATIC))
    {
        /* Only return value through resNum if the service type is
         * ALLOCATE... */
        if ((type == Rm_service_RESOURCE_ALLOCATE_INIT) ||
            (type == Rm_service_RESOURCE_ALLOCATE_USE))
        {
            *resNum = rmServiceResp.resourceBase;
        }     
        retVal = 1;
    }
    
    if (isAllocated)
    {
        if (rmServiceResp.resourceNumOwners == RM_RESOURCE_NUM_OWNERS_INVALID)
        {
            *isAllocated = 0;
        }
        else
        {
            /* QM would like to have the global allocation count, but RM doesn't provide that.
             * Most use cases rely on 0, 1, or 2 (last instance closed, first instance opened,
             * or subsequent instance opened.
             * The following logic will always return 0,1, or 2 correctly, but values >=2
             * only mean that this is not first open request
             */
            if ((rmServiceResp.resourceNumOwners == 1) && 
                (rmServiceResp.instAllocCount != RM_INST_ALLOC_COUNT_INVALID))
            {
                /* If this is only owner, return how many times it was opened on this instance; */
                *isAllocated = rmServiceResp.instAllocCount;
            }
            else
            {
                /* If there are multiple owners or no owners, return the owner count */
                *isAllocated = rmServiceResp.resourceNumOwners;
            }
            /* Error or denial occurred in RM but not total failure since refCount was returned */
            retVal = 1;
        }
    }   
    return (retVal);
}

/*
 *  @b Description
 *  @n  
 *      This function sends a service request to RM for each group supported by the QM
 *
 *  @param[in]  type
 *      Service request type
 *  @param[in]  nameOffset
 *      Offset to name pointer in Qmss_GlobalConfigGroupRm
 *  @param[out]  resNum
 *      Pointer to the resource value to check.  If the input value is
 *      unspecified the resource value returned by RM will be returned 
 *      through this parameter.
 *  @param[in]  resLen
 *      The resource length (starting from resNum) to check.
 *  @param[out] isAllocated
 *      Returns the reference count for the resource
 *  @param[in]  qmssGblCfgParams
 *      Structure describing groups
 *
 *  @retval
 *      1 - RM approved request.
 *      0 - RM denied request or experienced an error
 */
int Qmss_rmServiceGroups (Rm_ServiceHandle *rmService, Rm_ServiceType type, size_t nameOffset,
                          int32_t *resNum, uint32_t resLen, int32_t resAlign, uint8_t *isAllocated,
                          Qmss_GlobalConfigParams *qmssGblCfgParams)
{
    int qm, qm2;
    char *namePtr;

    /* Attempt to allocate control of all the QMs.  Right now both JOINT and SPLIT modes use all QMs */
    for (qm = 0; qm < qmssGblCfgParams->maxQueMgrGroups; qm++)
    {
        namePtr = ((char *)&qmssGblCfgParams->groupRm[qm]) + nameOffset;
        if ((!Qmss_rmService((Rm_ServiceHandle *)qmssGblCfgParams->qmRmServiceHandle, type,
                             namePtr, resNum, resLen, resAlign, isAllocated)))
        {
            /* Free any partially allocated resourcse */
            if (type != Rm_service_RESOURCE_FREE)
            {
                for (qm2 = qm - 1; qm2 >= 0; qm2--)
                {
                    namePtr = ((char *)&qmssGblCfgParams->groupRm[qm]) + nameOffset;
                    Qmss_rmService((Rm_ServiceHandle *)qmssGblCfgParams->qmRmServiceHandle, Rm_service_RESOURCE_FREE,
                                   namePtr, resNum, resLen, resAlign, isAllocated);
                }
            }

            return 0;
        }
    }

    return 1;
}

/**
 *  @b Description
 *  @n  
 *      This function initializes the Queue Manager subsystem low level driver
 *      This function is called once in the system to setup the queue manager 
 *      low level driver with information pertaining to linking RAM and descriptor 
 *      memory region.
 *
 *  @param[in]  initCfg
 *      Initialization structure that contains memory region configuration information.
 *      Linking RAM memory address should be a global address  
 *
 *  @param[in]  qmssGblCfgParams
 *      Initialization structure that contains the QMSS device specific information.
 *
 *  @post  
 *      QMSS instance is created.
 *
 *  @retval
 *      Success -   QMSS_SOK
 *  @retval
 *      Failure -   QMSS_INVALID_PARAM
 *  @retval
 *      Failure -   QMSS_RESOURCE_LINKING_RAM_INIT_DENIED
 *  @retval
 *      Failure -   QMSS_QM_CONTROL_DENIED
 */
Qmss_Result Qmss_init (Qmss_InitCfg *initCfg, Qmss_GlobalConfigParams *qmssGblCfgParams)
{
    Qmss_GlobalObj_Unpadded *qmss_objPtr = &qmssGObj.obj;
    uint32_t        index;
    void            *key;
    int             qGroup;
    uint32_t        qPerGroup;
    Qmss_LinkingRAM linkingRAMCfg;
    Qmss_Result     retVal = QMSS_SOK;
    Qmss_StartCfg   startCfg;
    
    if (initCfg == NULL)
    {
        return QMSS_INVALID_PARAM;
    }

    if (qmssGblCfgParams->maxPDSP > QMSS_MAX_PDSP) 
    {
        return QMSS_INVALID_PARAM;
    }

    if (qmssGblCfgParams->maxQueMgrGroups > QMSS_MAX_QMGR_GROUPS)
    {
        return QMSS_INVALID_PARAM;
    }

    for (index = 0; index < qmssGblCfgParams->maxPDSP; index++)
    {
        if ((initCfg->pdspFirmware[index].firmware != NULL) && (initCfg->pdspFirmware[index].size == 0))
            return QMSS_INVALID_PARAM;
    }

    if (qmssGblCfgParams->qmRmServiceHandle)
    {
        int32_t qmControl = 0; 

        /* Attempt to allocate control of both QMs.  Right now both JOINT and SPLIT modes use both QMs */
        if ((!Qmss_rmServiceGroups((Rm_ServiceHandle *)qmssGblCfgParams->qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_INIT,
                                   offsetof(Qmss_GlobalConfigGroupRm, rmQmControl), &qmControl, 1, 0, NULL, qmssGblCfgParams)))
        {
            return QMSS_QM_CONTROL_DENIED;
        }
    }     

    /* Check the max queue number configuration table(s) */
    qPerGroup = qmssGblCfgParams->maxQue / qmssGblCfgParams->maxQueMgrGroups;
    for (qGroup = 0; qGroup < qmssGblCfgParams->maxQueMgrGroups; qGroup++) 
    {
        for (index = 0; index < QMSS_MAX_QUEUE_TYPES; index++) 
        {
             Qmss_QueueType qType      = qmssGblCfgParams->maxQueueNum[qGroup][index].queueType;
             int32_t        startIndex = qmssGblCfgParams->maxQueueNum[qGroup][index].startIndex;
             /* Each row's type must match its index, or be GENERAL.
              * The special case for GENERAL allows the rest of the table to be filled
              * with additional GENERAL queues
              */
             if (qType != (Qmss_QueueType)index)
             {
                 if (index == 0) 
                 {
                     /* First entery type cannot be GENERAL */
                       return QMSS_INVALID_QUEUE_NUMBER_TAB;
                 }
                 if ((index > 0) && (qType != Qmss_QueueType_GENERAL_PURPOSE_QUEUE))
                 {
                     /* Extra GENERAL PURPOSE may be at end of table */
                       return QMSS_INVALID_QUEUE_NUMBER_TAB;
                 }
             }
             /* Check that the queue index matches the queue group */
             if (startIndex && ((startIndex < (qPerGroup * qGroup)) || 
                                (startIndex >= (qPerGroup * (qGroup + 1)))))
             {
                 return QMSS_INVALID_QUEUE_NUMBER_TAB;
             }
        }
    }

    /* Begin Critical Section before accessing shared resources. */
    key = Qmss_osalCsEnter ();

    /* Initialize private data structures */
    memset ((void *) qmssQueueFree, 0, sizeof (qmssQueueFree));
    Qmss_osalEndMemAccess ((void *) qmssQueueFree, sizeof (qmssQueueFree));
    for (qGroup = 0; qGroup < qmssGblCfgParams->maxQueMgrGroups; qGroup++) 
    {
        qmssGroupControl.groupFreeQueues[qGroup] = qPerGroup;
    }
    Qmss_osalEndMemAccess ((void *)&qmssGroupControl, sizeof (qmssGroupControl));

    memset (qmss_objPtr, 0, sizeof (Qmss_GlobalObj_Unpadded));
    for (index = 0; index < QMSS_MAX_MEM_REGIONS; index++)
        qmss_objPtr->descQueue[index] = -1;

    /* Store a local copy of the initial memory region configuration */
    memcpy ((void *) &qmss_objPtr->initCfg, initCfg, sizeof (Qmss_InitCfg));

    linkingRAMCfg.linkingRAM0Base = initCfg->linkingRAM0Base;
    linkingRAMCfg.linkingRAM0Size = initCfg->linkingRAM0Size;
    linkingRAMCfg.linkingRAM1Base = initCfg->linkingRAM1Base;
    linkingRAMCfg.maxDescNum      = initCfg->maxDescNum;

    /* Check the linking RAM config */
    if ((retVal = Qmss_internalCheckLinkingRAM (&linkingRAMCfg)) != QMSS_SOK) 
    {
       goto exitCs;
    }

    /* If split mode check the rest of the linking RAM */
    if (initCfg->mode == Qmss_Mode_SPLIT) 
    {
        for (qGroup = 1; qGroup < qmssGblCfgParams->maxQueMgrGroups; qGroup++) 
        {
            if ((retVal = Qmss_internalCheckLinkingRAM (initCfg->splitLinkingRAMs + qGroup)) != QMSS_SOK) 
            {
               goto exitCs;
            }
        }
    }


    /* Copy global configuration parameters */
    memcpy ((void *) &qmss_objPtr->qmssGblCfgParams, qmssGblCfgParams, sizeof (Qmss_GlobalConfigParams));
    /* Writeback now, as Qmss_startCfg will invalidate for external callability */
    Qmss_osalEndMemAccess ((void *) qmss_objPtr, sizeof (*qmss_objPtr));
    /* Mark global object valid valid */
    qmssGObjIsValid = 1;
    Qmss_osalEndMemAccess ((void *) &qmssGObjIsValid, sizeof (qmssGObjIsValid));

    startCfg.rmServiceHandle = qmssGblCfgParams->qmRmServiceHandle;
    Qmss_startCfg (&startCfg);

    if (qmssLObj.p.qmRmServiceHandle && (qmssGObj.obj.initCfg.qmssHwStatus != QMSS_HW_INIT_COMPLETE))
    {
        /* Check Linking RAM 0 and 1 control permissions */
        int32_t linkRamControl = 0; 
        /* Attempt to allocate control of all the QMs.  Right now both JOINT and SPLIT modes use all QMs */
        if ((!Qmss_rmServiceGroups((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_INIT, 
                                   offsetof(Qmss_GlobalConfigGroupRm,rmLinkRamControl),
                                   &linkRamControl, 1, 0, NULL, qmssGblCfgParams)))
        {
            /* Set the return value to signify setting base address registers was denied and set the
             * HW_INIT_COMPLETE to be true in the global config.  This signifies another entity has
             * configured the QMSS and some entries of the qmssGObj may not be valid, such as
             * the initCfg.maxDescNum field */
            retVal = QMSS_RESOURCE_LINKING_RAM_INIT_DENIED;
            qmssGObj.obj.initCfg.qmssHwStatus = QMSS_HW_INIT_COMPLETE;
        }
    }   

    if (qmssGObj.obj.initCfg.qmssHwStatus != QMSS_HW_INIT_COMPLETE)
    {
        /* Insert the linking RAM for QM 0 */
        Qmss_internalInsertLinkingRAM (0, &linkingRAMCfg, &qmss_objPtr->initCfg.linkingRAM0Size);
     
        /* Insert the linking RAM for other QMs in split mode */
        for (qGroup = 1; qGroup < qmssGblCfgParams->maxQueMgrGroups; qGroup++) 
        {
           Qmss_LinkingRAM *lram_p;
           uint32_t        *size_p;
           if (initCfg->mode == Qmss_Mode_SPLIT) 
           {
               /* Each group gets its own config */
               lram_p = initCfg->splitLinkingRAMs + qGroup - 1;
               size_p = &qmss_objPtr->initCfg.splitLinkingRAMs[qGroup - 1].linkingRAM0Size;
           }
           else
           {
               /* Same config into all groups */
               lram_p = &linkingRAMCfg;
               size_p = &qmss_objPtr->initCfg.linkingRAM0Size;
           }
           Qmss_internalInsertLinkingRAM (qGroup, lram_p, size_p);
        }
        /* Note: qmssHwStatus is NOT QMSS_HW_INIT_COMPLETE, this is used in Qmss_exit
         * to avoid freeing linking ram */
    }

    /* Download Firmware */
    for (index = 0; index < qmssGblCfgParams->maxPDSP; index++)
    {
        if (initCfg->pdspFirmware[index].firmware != NULL)
        {
            retVal = Qmss_internaldownloadFirmware (initCfg->pdspFirmware[index].pdspId, initCfg->pdspFirmware[index].firmware, 
                                    initCfg->pdspFirmware[index].size);
        }
    }

    /* Writeback Global Object - required because Qmss_internalInsertLinkingRAM can write to initCfg.linkingRAM0Size */
    Qmss_osalEndMemAccess ((void *) &qmssGObj.obj, sizeof (Qmss_GlobalObj_Unpadded));

exitCs:
    /* End Critical Section */
    Qmss_osalCsExit (key);
    return retVal;
}

/**
 *  @b Description
 *  @n  
 *      This function deinitializes the CPPI low level driver.
 *      The LLD is deinitialized only if all the queues (opened via this instance)
 *      are closed and regions inserted through this instance are removed.
 *      It is user responsibility to dispose of all descriptors added via this
 *      instance.
 *
 *  @pre  
 *      Qmss_init function should be called before calling this function.
 *      All descriptors added by this instance should be removed with Qmss_queueEmpty().
 *      All regions added by this instance shoudl be removed with Qmss_removeMemoryRegion().
 *
 *  @retval
 *      Success -   QMSS_SOK
 *  @retval
 *      Failure -   QMSS_NOT_INITIALIZED
 *                  QMSS_EXIT_QUEUES_OPEN
 *                  QMSS_EXIT_REGIONS_PRESENT
 *                  QMSS_QM_CONTROL_DENIED
 *                  QMSS_RESOURCE_LINKING_RAM_INIT_DENIED
 */
Qmss_Result Qmss_exit (void)
{
    Qmss_Result              retVal = QMSS_SOK;
    void                    *key;
    uint32_t                 qPerGroup;
    int                      qGroup;
    int                      pdsp;
    Qmss_GlobalConfigParams *qmssGblCfgParams;

    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;

    /* Begin Critical Section before accessing shared resources. */
    key = Qmss_osalCsEnter ();

    /* Invalidate Global Object */
    Qmss_osalBeginMemAccess ((void *) &qmssGObj.obj, sizeof (Qmss_GlobalObj_Unpadded));
    /* Invalidate group control */
    Qmss_osalBeginMemAccess ((void *) &qmssGroupControl, sizeof (qmssGroupControl));

    /* Check if all regions are removed */
    if (qmssGObj.obj.currDescCnt)
    {
        retVal = QMSS_EXIT_REGIONS_PRESENT;
        goto exitCs;
    }

    qmssGblCfgParams = &qmssGObj.obj.qmssGblCfgParams;

    /* Check if all queues are removed */
    qPerGroup = qmssGblCfgParams->maxQue / qmssGblCfgParams->maxQueMgrGroups;
    for (qGroup = 0; qGroup < qmssGblCfgParams->maxQueMgrGroups; qGroup++) 
    {
        if (qmssGroupControl.groupFreeQueues[qGroup] != qPerGroup)
        {
            retVal = QMSS_EXIT_QUEUES_OPEN;
            goto exitCs;
        }
    }

    if (qmssLObj.p.qmRmServiceHandle)
    {
        int32_t qmControl = 0; 
        int32_t linkRamControl = 0;

        /* Release the control */
        if ((!Qmss_rmServiceGroups((Rm_ServiceHandle *)qmssGblCfgParams->qmRmServiceHandle, Rm_service_RESOURCE_FREE,
                                   offsetof(Qmss_GlobalConfigGroupRm, rmQmControl), &qmControl, 1, 0, NULL, qmssGblCfgParams)))
        {
            /* Indicate RM wasn't able to free control */
            retVal = QMSS_QM_CONTROL_DENIED;
        }

        /* Release the linking RAM control only if we took it */
        if (qmssGObj.obj.initCfg.qmssHwStatus != QMSS_HW_INIT_COMPLETE)
        {
            if ((!Qmss_rmServiceGroups((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_FREE, 
                                       offsetof(Qmss_GlobalConfigGroupRm,rmLinkRamControl),
                                       &linkRamControl, 1, 0, NULL, qmssGblCfgParams)))
            {
                /* Indicate RM couldn't free linking RAM */
                retVal = QMSS_RESOURCE_LINKING_RAM_INIT_DENIED;
            }
        }

        /* Release the firmware if we took it */
        for (pdsp = 0; pdsp < qmssGObj.obj.qmssGblCfgParams.maxPDSP; pdsp++)
        {
            if (qmssGObj.obj.initCfg.pdspFirmware[pdsp].firmware != NULL)
            {
                /* Release firmware "use" */
                int32_t pdspIdNum = qmssGObj.obj.initCfg.pdspFirmware[pdsp].pdspId;
                if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_FREE, 
                                    qmssLObj.p.rmFirmwarePdsp, &pdspIdNum, 1, 0, NULL))
                {
                    retVal = QMSS_RESOURCE_ALLOCATE_USE_DENIED;
                }
                else
                {
                    /* Release firmware "init" */
                    if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_FREE, 
                                        qmssLObj.p.rmFirmwarePdsp, &pdspIdNum, 1, 0, NULL))
                    {
                        retVal = QMSS_RESOURCE_ALLOCATE_INIT_DENIED;
                    }
                }
            }
        }
    }

    /* Clear local object */
    qmssLObjIsValid = 0;

exitCs:
    Qmss_osalCsExit(key);

    return retVal;
}

/**
 *  @b Description
 *  @n  
 *      This function initializes the Queue Manager SubsSystem local object blocks in memory 
 *      local to each core and initializes any start parameters like the RM instance. 
 *      
 *      This function must be called at least once for every core.  It can either be called directly
 *      through the application if RM is to be used or it will be called via Qmss_start if RM is
 *      not to be used.
 *
 *  @pre  
 *      Qmss_init function should be called before calling this function.
 *
 *  @retval
 *      Success -   QMSS_SOK
 *  @retval
 *      Failure -   QMSS_NOT_INITIALIZED
 */
Qmss_Result Qmss_startCfg (Qmss_StartCfg *startCfg)
{
    if (qmssLObjIsValid == 0)
    {
        /* Invalidate Global Object Valid */
        Qmss_osalBeginMemAccess ((void *) &qmssGObjIsValid, sizeof (qmssGObjIsValid));
        if (qmssGObjIsValid)
        {
            memset ((void *) &qmssLObj.p, 0, sizeof (Qmss_GlobalConfigParams));
            /* Invalidate Global Object */
            Qmss_osalBeginMemAccess ((void *) &qmssGObj.obj, sizeof (Qmss_GlobalObj_Unpadded));
            qmssLObj.p = qmssGObj.obj.qmssGblCfgParams;
            qmssLObj.p.qmRmServiceHandle = startCfg->rmServiceHandle;
            qmssLObjIsValid = 1;
        }
        else
        {
            return QMSS_NOT_INITIALIZED;
        }
    }

    return QMSS_SOK;
}

/**
 *  @b Description
 *  @n  
 *      This function initializes the Queue Manager SubsSystem local object blocks in memory 
 *      local to each core. 
 *      
 *      This function must be called atleast once for every core.
 *
 *  @pre  
 *      Qmss_init function should be called before calling this function.
 *
 *  @retval
 *      Success -   QMSS_SOK
 *  @retval
 *      Failure -   QMSS_NOT_INITIALIZED
 */
Qmss_Result Qmss_start (void)
{
    Qmss_StartCfg startCfg;

    /* For backwards compatability make the RM service handle NULL */
    startCfg.rmServiceHandle = NULL;
    
    return(Qmss_startCfg(&startCfg));
}

/**
 *  @b Description
 *  @n  
 *      This function is used to obtain the QMSS memory region configuration.
 *      The function returns the descriptor base address, descriptor size and
 *      the number of descriptors for each configured memory region.
 *
 *  @param[out]  memRegInfo
 *      Pointer to the memory region configuration structure allocated by caller.
 * 
 *  @pre  
 *      Qmss_init function should be called before calling this function.
 *       
 *  @retval
 *      Success -   QMSS_SOK
 *  @retval
 *      Failure -   QMSS_INVALID_PARAM
 *  @retval
 *      Failure -   QMSS_NOT_INITIALIZED
 */
Qmss_Result Qmss_getMemoryRegionCfg (Qmss_MemRegCfg *memRegInfo)
{
    uint32_t          i;            
    uint32_t          qGroup;

    if (memRegInfo == NULL)
        return QMSS_INVALID_PARAM;

    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;

    /* No RM check since readonly operation */

    memset ((void *) memRegInfo, 0, sizeof (Qmss_MemRegCfg));

    /* Invalidate Global Object */
    Qmss_osalBeginMemAccess ((void *) &qmssGObj.obj, sizeof (Qmss_GlobalObj_Unpadded));

    /* Chose which group based on split/joint mode */
    if (qmssGObj.obj.mode == Qmss_Mode_SPLIT) 
    {
        /* In split mode, the config selects the group */
        qGroup = memRegInfo->queueGroup;
    } 
    else
    {
        /* In joint mode, the LLD tracks configuration use group 0, but the
         * configuration is inserted into all of the group registers
         */
        qGroup = 0;
    }

    if (qGroup > qmssGObj.obj.qmssGblCfgParams.maxQueMgrGroups) 
    {
        return QMSS_INVALID_PARAM;
    }

    for (i = 0; i < QMSS_MAX_MEM_REGIONS; i++)
    {
        memRegInfo->memRegInfo[i] = qmssGObj.obj.memRegInfo[qGroup][i];
    } 
    memRegInfo->currDescCnt = qmssGObj.obj.currDescCnt; 
    return QMSS_SOK;
}

/**
 *  @b Description
 *  @n  
 *      This function is used to configure memory region at runtime.
 *      The function configures specified memory region with descriptor base address, 
 *      descriptor size and the number of descriptors.
 *
 *  @param[out]  memRegCfg
 *      Pointer to the memory region configuration structure allocated by caller.
 *      Descriptor memory address should be a global address
 * 
 *  @pre  
 *      Qmss_init function should be called before calling this function.
 *      Memory Region specified should not have been configured before.
 *       
 *  @retval
 *      Success -   Inserted Memory region index. Range is 0 to 63. 
 *  @retval
 *      Failure -   QMSS_INVALID_PARAM
 *  @retval
 *      Failure -   QMSS_NOT_INITIALIZED
 *  @retval
 *      Failure -   QMSS_MEMREGION_ALREADY_INITIALIZED
 *  @retval
 *      Failure -   QMSS_MEMREGION_INVALID_PARAM
 *  @retval
 *      Failure -   QMSS_MAX_DESCRIPTORS_CONFIGURED
 *  @retval
 *      Failure -   QMSS_MEMREGION_INVALID_INDEX
 *  @retval
 *      Failure -   QMSS_MEMREGION_OVERLAP
 *  @retval
 *      Failure -   QMSS_QUEUE_OPEN_ERROR
 *  @retval
 *      Failure -   QMSS_RESOURCE_MEMORY_REGION_INIT_DENIED 
 *  @retval
 *      Failure -   QMSS_RESOURCE_LINKING_RAM_INIT_DENIED
 */
Qmss_Result Qmss_insertMemoryRegion (Qmss_MemRegInfo *memRegCfg)
{
    uint32_t          count, powRegSize, reg = 0;
    uint32_t          i, nextIndex; 
    uint32_t          descNum, oneBits;
    uint8_t           *descPtr, isAllocated;
    int32_t           queueHnd, index = -1, prevIndex, startIndex;
    void             *key;
    uint32_t         *descBasePhy;
    Qmss_Result       result;
    uint8_t          *newRegBegAddr, *newRegEndAddr;
    uint32_t          newRegBegIndex, newRegEndIndex;
    uint32_t          qGroup = 0;
    /* For installing same config over all queue groups */
    uint32_t          firstQGroup, regQGroup, lastQGroup; 
    int               rmMemRegionTaken = 0, rmStartIdxTaken = 0, hwProgrammed = 0;

    if (memRegCfg == NULL)
        return QMSS_INVALID_PARAM;

    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;

    /* Descriptor size should be a multiple of 16 */
    if ((memRegCfg->descSize % 16) != 0)
    {
        return QMSS_INVALID_PARAM;
    }

    /* Number of descriptors should be a power of 2 >= 2^5.  The upper limit is checked
     * below inside Qmss_osalCsEnter() */
    descNum = memRegCfg->descNum;
    if (descNum < 32)
    {
        return QMSS_INVALID_PARAM;
    }

    /* count the # of 1 bits, in order to see if its a power of 2 */
    for (oneBits = 0, descNum = memRegCfg->descNum; descNum; descNum >>=1)
    {
        if (descNum & 1) {
            oneBits++;
        }
    }

    /* Check if # of descriptors is a power of 2 */
    if (oneBits != 1) {
        return QMSS_INVALID_PARAM;
    }

    descBasePhy = (uint32_t*)Qmss_osalVirtToPhy((void *)(memRegCfg->descBase));
    if (!descBasePhy) {
        return QMSS_INVALID_PARAM;
    }

    /* The region base address must be aligned to 16 bytes.  It is suggested
     * to align to 64 bytes if the descriptor size > 32 bytes for performance
     * reasons.
     */
    if (((uint32_t)descBasePhy) & 0xf) {
        return QMSS_INVALID_PARAM;
    }

    /* Begin Critical Section before accessing shared resources. */
    key = Qmss_osalCsEnter ();
    
    /* Invalidate Global Object */
    Qmss_osalBeginMemAccess ((void *) &qmssGObj.obj, sizeof (Qmss_GlobalObj_Unpadded));
 
    /* Choose which group based on split/joint mode */
    if (qmssGObj.obj.mode == Qmss_Mode_SPLIT) 
    {
        /* In split mode, the config selects the group */
        qGroup = memRegCfg->queueGroup;
    } 
    else
    {
        /* In joint mode, the LLD tracks configuration use group 0, but the
         * configuration is inserted into all of the group registers
         */
        qGroup = 0;
    }
    if (qGroup > qmssGObj.obj.qmssGblCfgParams.maxQueMgrGroups) 
    {
        result = QMSS_INVALID_PARAM;
        goto exitCs;
    }

    /* Select correct memory region */

    /* Check if maximum descriptors already configured */
    if (qmssGObj.obj.currDescCnt + memRegCfg->descNum > qmssGObj.obj.initCfg.maxDescNum)
    {
        result = QMSS_MAX_DESCRIPTORS_CONFIGURED;
        goto exitCs; /* common exit to close CsEnter and MemAccess */
    }

    /* Memory region to configure is not specified. Find the next available one */
    if (memRegCfg->memRegion == Qmss_MemRegion_MEMORY_REGION_NOT_SPECIFIED)
    {
        if (qmssLObj.p.qmRmServiceHandle)
        {
            if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_INIT,
                                qmssLObj.p.groupRm[qGroup].rmMemRegion, &(memRegCfg->memRegion), 1, 0, NULL))
            {
                result = QMSS_RESOURCE_MEM_REGION_INIT_DENIED;
                goto exitCs;
            }
            rmMemRegionTaken = 1;
            index = memRegCfg->memRegion;
        }
        else
        {
            for (i = 0; i < QMSS_MAX_MEM_REGIONS; i++)
            {
                if (qmssGObj.obj.memRegInfo[qGroup][i].descNum == 0)
                {
                    index = i;
                    break;                    
                }
            }
        }
        if (qmssLObj.p.qmRmServiceHandle) {
            startIndex = QMSS_PARAM_NOT_SPECIFIED;
        }
        else {
            if (index == 0)
                startIndex = 0;
            else
                startIndex = qmssGObj.obj.memRegInfo[qGroup][index - 1].descNum + qmssGObj.obj.memRegInfo[qGroup][index - 1].startIndex;      
        }
    }
    else
    {
        if (qmssLObj.p.qmRmServiceHandle)
        {
            if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_INIT, 
                                qmssLObj.p.groupRm[qGroup].rmMemRegion, &(memRegCfg->memRegion), 1, 0, NULL))
            {
                result = QMSS_RESOURCE_MEM_REGION_INIT_DENIED;
                goto exitCs;
            }
            rmMemRegionTaken = 1;
        }
        index = memRegCfg->memRegion;
        startIndex = memRegCfg->startIndex;
    }

    if (qmssLObj.p.qmRmServiceHandle)
    {
        if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_INIT,
                            qmssLObj.p.groupRm[qGroup].rmLinkRam, &startIndex, memRegCfg->descNum, 0, NULL))
        {
            result = QMSS_RESOURCE_LINKING_RAM_INIT_DENIED;
            goto exitCs;
        }
        rmStartIdxTaken = 1;
    }
    
    /* Invalid memory region index */
    if (index < 0 || index >= QMSS_MAX_MEM_REGIONS)
    {
        result = QMSS_MEMREGION_INVALID_INDEX;
        goto exitCs; /* common exit to close CsEnter and MemAccess */
    }

    /* If specified memory region is already configured don't overwrite that configuration */
    if (qmssGObj.obj.memRegInfo[qGroup][index].descNum != 0)
    {
        result = QMSS_MEMREGION_ALREADY_INITIALIZED;
        goto exitCs; /* common exit to close CsEnter and MemAccess */
    }

    /* startIndex should be a multiple of 32 */
    if ((startIndex % 32) != 0)
    {
        result = QMSS_INVALID_PARAM;
        goto exitCs; /* common exit to close CsEnter and MemAccess */
    }
   
    /* Compute boundaries of the new region */
    newRegBegAddr  = (uint8_t *)Qmss_osalVirtToPhy((void *)(memRegCfg->descBase));
    newRegEndAddr  = newRegBegAddr + memRegCfg->descNum * memRegCfg->descSize;
    newRegBegIndex = startIndex;
    newRegEndIndex = newRegBegIndex + memRegCfg->descNum;

    /* Check if the new region overlaps any existing regions.  This is done 
     * with a brute force search that doesn't depend on ordering
     *
     * This allows overlap check to be kept if future hardware relaxes the
     * ordering requirements.
     */
    for (i = 0; i < QMSS_MAX_MEM_REGIONS; i++) 
    {
        Qmss_MemRegInfo *checkReg = &qmssGObj.obj.memRegInfo[qGroup][i];
        uint8_t *checkRegBegAddr  = 0;
        uint8_t *checkRegEndAddr  = 0;
        uint32_t checkRegBegIndex = checkReg->startIndex;
        uint32_t checkRegEndIndex = checkRegBegIndex + checkReg->descNum;
        
        if(checkReg->descBase)
        {
            /* Address translation if active */
            checkRegBegAddr = (uint8_t *)Qmss_osalVirtToPhy((void *)(checkReg->descBase));
        }
        checkRegEndAddr = checkRegBegAddr + (checkReg->descNum * checkReg->descSize);

        if (i == index) 
        {
            continue; /* don't test against myself */
        }

        if (! checkRegBegAddr) {
            continue; /* Check region hasn't been initialized */
        }

        /* Check the physical addresses for overlap */
        /* A) Is the new region's start address overlapping the check region? */
        if ((newRegBegAddr >= checkRegBegAddr) && 
            (newRegBegAddr < checkRegEndAddr)) 
        {   /* The new region starts inside the check region */
            result = QMSS_MEMREGION_OVERLAP;
            goto exitCs; /* common exit to close CsEnter and MemAccess */
        }
        /* B) Is the new region's end address overlapping the check region? */
        if ((newRegEndAddr > checkRegBegAddr) && 
            (newRegEndAddr <= checkRegEndAddr)) 
        {   /* The new region ends inside the check region */
            result = QMSS_MEMREGION_OVERLAP;
            goto exitCs; /* common exit to close CsEnter and MemAccess */
        }
        /* C) Is the check region completely inside the new region? */
        if ((checkRegBegAddr >= newRegBegAddr) && 
            (checkRegEndAddr <= newRegEndAddr)) 
        {   /* The check region is contained within the new region */
            result = QMSS_MEMREGION_OVERLAP;
            goto exitCs; /* common exit to close CsEnter and MemAccess */
        }
        /* It is not necessary to check if the new region is completely inside
         * the check region because that is covered by A and B above since both
         * its start and end addresses are inside the check region.
         */

        /* Check the linking RAM index for overlap */
        /* Is the new region's start index overlapping the check region? */
        if ((newRegBegIndex >= checkRegBegIndex) && 
            (newRegBegIndex < checkRegEndIndex)) 
        {   /* The new region starts inside the check region */
            result = QMSS_MEMREGION_OVERLAP;
            goto exitCs; /* common exit to close CsEnter and MemAccess */
        }
        /* Is the new region's end index overlapping the check region? */
        if ((newRegEndIndex > checkRegBegIndex) && 
            (newRegEndIndex <= checkRegEndIndex)) 
        {   /* The new region ends inside the check region */
            result = QMSS_MEMREGION_OVERLAP;
            goto exitCs; /* common exit to close CsEnter and MemAccess */
        }
        /* Is the check region completely inside the new region? */
        if ((checkRegBegIndex >= newRegBegIndex) && 
            (checkRegEndIndex <= newRegEndIndex)) 
        {   /* The check region is contained within the new region */
            result = QMSS_MEMREGION_OVERLAP;
            goto exitCs; /* common exit to close CsEnter and MemAccess */
        }
        /* It is not necessary to check if the new region is completely inside
         * the check region because that is covered by A and B above since both
         * its start and end addresses are inside the check region.
         */
    }

    /* Some devices don't constrain the order of memory regions */
    if (qmssLObj.p.orderedMemReg)
    {
        /* Find next and previous initialized regions to check for ordering */
        for (prevIndex = index-1; 
             (prevIndex >= 0 && (qmssGObj.obj.memRegInfo[qGroup][prevIndex].descNum == 0)); 
             prevIndex--);
        for (nextIndex = index+1; 
             ((nextIndex <= (QMSS_MAX_MEM_REGIONS-1) && 
              (qmssGObj.obj.memRegInfo[qGroup][nextIndex].descNum == 0))); 
             nextIndex++);

        /* If there is a previous region make sure its index and its physical 
         * address is lower than the new region's.
         *
         * It is assumed that the index and address do not overlap the previous 
         * region since that is already verified above.
         */
        if (prevIndex >= 0) 
        {
            Qmss_MemRegInfo *prevReg       = &qmssGObj.obj.memRegInfo[qGroup][prevIndex];

            /* Check that the linking RAM index is strictly increasing */
            if (prevReg->startIndex >= newRegBegIndex) {
                result = QMSS_MEMREGION_ORDERING;
                goto exitCs; /* common exit to close CsEnter and MemAccess */
            }

            /* Check that the physical memory address is strictly increasing */
            if (((uint8_t *)Qmss_osalVirtToPhy((void *)(prevReg->descBase))) >= newRegBegAddr) {
                result = QMSS_MEMREGION_ORDERING;
                goto exitCs; /* common exit to close CsEnter and MemAccess */
            }
        }

        /* If there is a next region make sure its index and its physical 
         * address is lower than the new region's.
         *
         * It is assumed that the index and address do not overlap the next 
         * region since that is already verified above.
         */
        if (nextIndex < QMSS_MAX_MEM_REGIONS) 
        {
            Qmss_MemRegInfo *nextReg       = &qmssGObj.obj.memRegInfo[qGroup][nextIndex];

            /* Check that the linking RAM index is strictly increasing */
            if (nextReg->startIndex <= newRegBegIndex) {
                result = QMSS_MEMREGION_ORDERING;
                goto exitCs; /* common exit to close CsEnter and MemAccess */
            }

            /* Check that the physical memory address is strictly increasing */
            if (((uint8_t *)Qmss_osalVirtToPhy((void *)(nextReg->descBase))) <= newRegBegAddr) {
                result = QMSS_MEMREGION_ORDERING;
                goto exitCs; /* common exit to close CsEnter and MemAccess */
            }
        }
    }

    /* If the qmssGObj is valid (i.e. HW INIT was performed in the DSP subsystem
      * check that there is sufficient linking RAM */
    if (qmssGObj.obj.initCfg.qmssHwStatus != QMSS_HW_INIT_COMPLETE)
    {
        if (newRegEndIndex > qmssGObj.obj.initCfg.maxDescNum)
        {
            result = QMSS_INVALID_PARAM;
            goto exitCs; /* common exit to close CsEnter and MemAccess */
        }
    }

    /* Configure the QMSS memory regions */
    if (qmssGObj.obj.mode == Qmss_Mode_SPLIT) 
    {
        /* Apply configuration to selected group */
        firstQGroup = qGroup;
        lastQGroup  = qGroup;
    }
    else 
    {
        /* Apply configuration to all groups */
        firstQGroup = 0;
        lastQGroup  = qmssGObj.obj.qmssGblCfgParams.maxQueMgrGroups - 1;
    }
    
    for (regQGroup = firstQGroup; regQGroup <= lastQGroup; regQGroup++) 
    {
        qmssLObj.p.groupRegs[regQGroup].qmDescReg->MEMORY_REGION_BASE_ADDRESS_GROUP[index].MEMORY_REGION_BASE_ADDRESS_REG = (uint32_t)descBasePhy;
    
        CSL_FINS (qmssLObj.p.groupRegs[regQGroup].qmDescReg->MEMORY_REGION_BASE_ADDRESS_GROUP[index].MEMORY_REGION_START_INDEX_REG, 
            QM_DESCRIPTOR_REGION_CONFIG_MEMORY_REGION_START_INDEX_REG_START_INDEX, startIndex);

        CSL_FINS (reg, QM_DESCRIPTOR_REGION_CONFIG_MEMORY_REGION_DESCRIPTOR_SETUP_REG_DESC_SIZE, 
            ((memRegCfg->descSize / 16) - 1));

        for (powRegSize = 0; (32UL << powRegSize) < memRegCfg->descNum; powRegSize++);

        CSL_FINS (reg, QM_DESCRIPTOR_REGION_CONFIG_MEMORY_REGION_DESCRIPTOR_SETUP_REG_REG_SIZE, powRegSize);

        qmssLObj.p.groupRegs[regQGroup].qmDescReg->MEMORY_REGION_BASE_ADDRESS_GROUP[index].MEMORY_REGION_DESCRIPTOR_SETUP_REG = reg; 
    }

    hwProgrammed = 1;

    /* Check if LLD is supposed to manage the descriptors, if so slice up the 
     * descriptors and push to queues. else do nothing the caller will manage 
     * the descriptors in memory region */
    if (memRegCfg->manageDescFlag)
    {
        /* Split and push to general purpose queues */

        /* Open general purpose queue */
        queueHnd = Qmss_internalQueueOpen (0, qGroup, Qmss_QueueType_GENERAL_PURPOSE_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
        if (queueHnd < 0)
        {            
            result = QMSS_QUEUE_OPEN_ERROR;
            goto exitCs; /* common exit to close osalCsEnter and osalBeginMemAccess */
        }
            
        /* Store the handle */
        qmssGObj.obj.descQueue[index] = queueHnd;
        
        descPtr = (uint8_t *) memRegCfg->descBase;
        for (count = 0; count < memRegCfg->descNum; count ++) 
        {
            /* TODO QueueProxy is not working add this later 
             * Qmss_queuePushDesc (queueHnd, (uint32_t *)descPtr, initCfgPtr->memRegionInfo[index].descSize); */
        
            Qmss_queuePushDesc (queueHnd, (uint32_t *) descPtr);
            descPtr = descPtr + memRegCfg->descSize;
        }
    }

    /* Store in internal data structures */
    qmssGObj.obj.memRegInfo[qGroup][index]            = *memRegCfg;
    qmssGObj.obj.memRegInfo[qGroup][index].startIndex = startIndex;
    qmssGObj.obj.memRegInfo[qGroup][index].memRegion  = (Qmss_MemRegion) index;
    qmssGObj.obj.currDescCnt += memRegCfg->descNum;

    /* Return value */
    result = index;
    
exitCs:
    /* Writeback Global Object */
    Qmss_osalEndMemAccess ((void *) &qmssGObj.obj, sizeof (Qmss_GlobalObj_Unpadded));

    /* RM Error cleanup */
    if (qmssLObj.p.qmRmServiceHandle && (result < QMSS_SOK) && (! hwProgrammed))
    {
        if (rmMemRegionTaken)
        {
            if (! Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_FREE, 
                                 qmssLObj.p.groupRm[qGroup].rmMemRegion, &(memRegCfg->memRegion), 1, 0, NULL))
            {
                result = QMSS_RESOURCE_FREE_DENIED; /* double error */
            }
        }
        if (rmStartIdxTaken)
        {
            if (! Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_FREE,
                                 qmssLObj.p.groupRm[qGroup].rmLinkRam, &startIndex, memRegCfg->descNum, 0, NULL))
            {
                result = QMSS_RESOURCE_FREE_DENIED; /* double error */
            }
        }
    }

    /* End Critical Section */
    Qmss_osalCsExit (key);
    return result;
}

/**
 *  @b Description
 *  @n  
 *      This function is used to deconfigure memory region at runtime.
 *      The hardware region is set to NULL, and any associated queue
 *      is empted and closed.
 *
 *  @param[in] region
 *      Memory region to clear (return value of Qmss_insertMemoryRegion)
 *
 *  @param[in] qGroup
 *      Queue group to clear region from (only used in split mode)
 * 
 *  @pre  
 *      All descriptors must be discarded by application from all queues
 *      before using this function, or it will leave hw in unknown state.
 *       
 *  @retval
 *      Success -   QMSS_SOK
 *  @retval
 *      Failure -   QMSS_INVALID_PARAM, QMSS_NOT_INITIALIZED
 */
extern Qmss_Result Qmss_removeMemoryRegion (int32_t region, uint32_t qGroup)
{
    void             *key;
    /* For installing same config over all queue groups */
    uint32_t          firstQGroup, regQGroup, lastQGroup; 
    Qmss_QueueHnd     freeQ = -1;
    Qmss_Result       retVal = QMSS_SOK;
    Qmss_MemRegInfo  *pRegCfg;

    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;

    /* Chose which group based on split/joint mode */
    if (qmssGObj.obj.mode != Qmss_Mode_SPLIT) 
    {
        /* In joint mode, the LLD tracks configuration use group 0, but the
         * configuration is inserted into all of the group registers
         */
        qGroup      = 0;
        firstQGroup = 0;
        lastQGroup  = qmssGObj.obj.qmssGblCfgParams.maxQueMgrGroups - 1;
    }
    else
    {
        firstQGroup = qGroup;
        lastQGroup  = qGroup;
    }

    if (qGroup > qmssGObj.obj.qmssGblCfgParams.maxQueMgrGroups) 
    {
        return QMSS_INVALID_PARAM;
    }

    /* Begin Critical Section before accessing shared resources. */
    key = Qmss_osalCsEnter ();

    /* Invalidate Global Object */
    Qmss_osalBeginMemAccess ((void *) &qmssGObj.obj, sizeof (Qmss_GlobalObj_Unpadded));

    pRegCfg = &qmssGObj.obj.memRegInfo[qGroup][region];
    
    if (qmssLObj.p.qmRmServiceHandle)
    {
        int32_t  linkRamStartIndex = pRegCfg->startIndex;
        uint32_t linkRamDescNum = pRegCfg->descNum;
            
        if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_FREE,
                            qmssLObj.p.groupRm[qGroup].rmMemRegion, &region, 1, 0, NULL))
        {
            retVal = QMSS_RESOURCE_FREE_DENIED;
            goto exitCs;
        }

        if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_FREE,
                            qmssLObj.p.groupRm[qGroup].rmLinkRam, &linkRamStartIndex, linkRamDescNum, 0, NULL))
        {
            retVal = QMSS_RESOURCE_FREE_DENIED;
            goto exitCs;
        }        
    }

    /* If the LLD is keeping a free descriptor queue, clear it and close it */
    if (pRegCfg->manageDescFlag)
    {
        freeQ = qmssGObj.obj.descQueue[region];
        qmssGObj.obj.descQueue[region] = -1;
        /* trash the descriptors */
        Qmss_queueEmpty (freeQ);
        retVal = Qmss_internalQueueClose (freeQ);
    }

    /* Clear the config */
    qmssGObj.obj.currDescCnt -= pRegCfg->descNum;
    memset (pRegCfg, 0, sizeof(*pRegCfg));

    /* Writeback */
    Qmss_osalEndMemAccess ((void *) &qmssGObj.obj, sizeof (Qmss_GlobalObj_Unpadded));
 
    /* Clear the registers */
    for (regQGroup = firstQGroup; regQGroup <= lastQGroup; regQGroup++) 
    {
        qmssLObj.p.groupRegs[regQGroup].qmDescReg->MEMORY_REGION_BASE_ADDRESS_GROUP[region].MEMORY_REGION_BASE_ADDRESS_REG = (uint32_t)0;
        qmssLObj.p.groupRegs[regQGroup].qmDescReg->MEMORY_REGION_BASE_ADDRESS_GROUP[region].MEMORY_REGION_DESCRIPTOR_SETUP_REG = 0; 
    }

exitCs:
    /* End Critical Section */
    Qmss_osalCsExit (key);

    return retVal;
}

/**
 *  @b Description
 *  @n  
 *      This function is used to obtain previously allocated descriptors.
 *      The function opens a destination queue as specified in the input parameters 
 *      and pushes the requested number of descriptors if available onto it. 
 *
 *  @param[in]  descCfg
 *      Specifies the number of descriptors, memory region from where it should be allocated 
 *      and the destination queue to push to.
 *      
 *  @param[out]  numAllocated
 *      Number of descriptors actually allocated.
 * 
 *  @pre  
 *      Qmss_init function should be called before calling this function.
 *       
 *  @retval
 *      Success -   Destination queue handle on which the allocated descriptors are stored. 
 *                  The queue must be closed by the caller using Qmss_queueClose() when done.

 *  @retval
 *      Failure -   QMSS_NOT_INITIALIZED
 *  @retval
 *      Failure -   QMSS_INVALID_PARAM
 *  @retval
 *      Failure -   QMSS_MEMREGION_NOT_INITIALIZED
 *  @retval
 *      Failure -   QMSS_QUEUE_OPEN_ERROR
 *  @retval
 *      Failure -   QMSS_RESOURCE_MEM_REGION_USE_DENIED
 */
Qmss_QueueHnd Qmss_initDescriptor (Qmss_DescCfg *descCfg, uint32_t *numAllocated)
{
    uint32_t              count;
    uint32_t              *descPtr;
    uint8_t               isAllocated;
    int32_t               srcQueNum, srcQueHnd, destQueHnd;
    uint32_t              qGroup;

    /* Get the descriptor from the source queue */
    
    if (qmssLObjIsValid == 0)
    {
        return QMSS_NOT_INITIALIZED;
    }

    if (descCfg == NULL)
    {
        return QMSS_INVALID_PARAM;
    }

    /* Find the source queue given the memory region */
    if ((srcQueNum = Qmss_getMemRegQueueHandle (descCfg->memRegion)) < 0)
    {
        return QMSS_MEMREGION_NOT_INITIALIZED;
    }

    if (qmssLObj.mode == Qmss_Mode_SPLIT) 
    {
        /* In split mode, the config selects the group */
        qGroup = descCfg->queueGroup;
    } 
    else
    {
        /* In joint mode, the LLD tracks configuration use group 0, but the
         * configuration is inserted into all of the group registers
         */
        qGroup = 0;
    }

    if (qGroup > qmssGObj.obj.qmssGblCfgParams.maxQueMgrGroups) 
    {
        return QMSS_INVALID_PARAM;
    }


    /* Find the descriptor size in the given the memory region */
    if ((Qmss_getMemRegDescSize (descCfg->memRegion, qGroup)) == 0)
    {
        return QMSS_MEMREGION_NOT_INITIALIZED;
    }

    /* Open destination queue specified in the configuration */
    if ((destQueHnd = Qmss_queueOpenInGroup (qGroup, descCfg->queueType, descCfg->destQueueNum, &isAllocated)) < 0)
    {
        return QMSS_QUEUE_OPEN_ERROR;
    }

    /* Open general purpose queue on which descriptors are stored */
    if ((srcQueHnd = Qmss_queueOpenInGroup (qGroup, Qmss_QueueType_GENERAL_PURPOSE_QUEUE, srcQueNum, &isAllocated)) < 0)
    {
        return QMSS_QUEUE_OPEN_ERROR;
    }
    *numAllocated = 0;
    for (count = 0; count < descCfg->descNum; count++)
    {
        /* Pop descriptor from source queue */
        if ((descPtr = Qmss_queuePop (srcQueHnd)) == NULL)
        {
            break;
        }

        /* Push descriptor to the specified destination queue */
        
        /* TODO QueueProxy is not working add this later 
         * Qmss_queuePushDesc (destQueHnd, descPtr, descSize); */
        
        Qmss_queuePushDesc (destQueHnd, descPtr);
        *numAllocated += 1;
    }

    Qmss_queueClose (srcQueHnd);

    return destQueHnd;
}

/**
 *  @b Description
 *  @n  
 *      This function opens the requested queue.
 *      A queue can be opened in two ways:
 *          1) If "queNum" is set to QMSS_PARAM_NOT_SPECIFIED, then a new
 *          available queue of type "queType" is allocated.
 *          2) If "queNum" is a valid queue number i.e., >= 0, 
 *               then a queue is allocated if free 
 *               else
 *               a handle to a previously opened queue is returned.
 *               In this case "queType" parameter is not used.
 *
 *  @param[in]  queGroup
 *      Specifies the queue manager group to find a queue.  This must be specified
 *      in SPLIT mode.  It may be specified (not QMSS_PARAM_NOT_SPECIFIED) in JOINT
 *      mode to allocate a queue in the same group as another queue.
 *
 *  @param[in]  queType
 *      Specifies the type of queue that should be opened. 
 * 
 *  @param[in]  queNum
 *      Specifies the queue number that should be opened.
 *      
 *  @param[out]  isAllocated
 *      Indicates whether the requested queue is a new queue allocation(1).
 *      or was already allocated. If the queue was previously allocated this parameter returns 
 *      the reference count.
 *
 *  @retval
 *      Success -   Queue Handle. Used as an input parameter for operation on this queue.
 *  @retval
 *      Failure -   QMSS_RESOURCE_ALLOCATE_USE_DENIED 
 *  @retval
 *      Failure -   <0
 */
Qmss_QueueHnd Qmss_queueOpenInGroup (int32_t queGroup, Qmss_QueueType queType, int32_t queNum, uint8_t *isAllocated)
{
    Qmss_QueueHnd   result;
    void            *key;
    uint32_t        openQGroup;
    int             qGroupIsSpecified;

    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;

    /* Begin Critical Section before accessing shared resources. */
    key = Qmss_osalCsEnter ();

    /* Convert signed PARAM_NOT_SPECIFIED into two arguments.  0 is legal queue group. */
    openQGroup        = (uint32_t)queGroup;
    qGroupIsSpecified = 1;
    if (queGroup == QMSS_PARAM_NOT_SPECIFIED) 
    {
        openQGroup = 0;
        qGroupIsSpecified = 0;
    }
    result = Qmss_internalQueueOpen (qGroupIsSpecified, openQGroup, queType, queNum, isAllocated);

    /* End Critical Section */   
    Qmss_osalCsExit (key);

    return (Qmss_QueueHnd) result;
}
/**
 *  @b Description
 *  @n  
 *      This function opens the requested queue from queue group 0.  For
 *      additional details, see @ref Qmss_queueOpenInGroup.  
 *
 *      When operating in SPLIT mode, this function will allocate the queue
 *      from group 0.  To use other groups, must use @ref Qmss_queueOpenInGroup.
 *
 *      When operating in JOINT mode, this function provides load balancing
 *      using the configured load balancing mode.
 *
 *  @retval
 *      Success -   Queue Handle. Used as an input parameter for operation on this queue.
 *  @retval
 *      Failure -   <0
 */

Qmss_QueueHnd Qmss_queueOpen (Qmss_QueueType queType, int32_t queNum, uint8_t *isAllocated)
{
    return Qmss_queueOpenInGroup (QMSS_PARAM_NOT_SPECIFIED, queType, queNum, isAllocated);
}

/**
 *  @b Description
 *  @n  
 *      This function opens the requested queue if a free queue is avialable within the given start and end range. 
 *
 *  @param[in]  startQueNum
 *      Specifies the start of the queue number range.
 * 
 *  @param[in]  endQueNum
 *      Specifies the end of the queue number range.
 * 
 *  @param[out]  isAllocated
 *      Indicates whether the requested queue is a new queue allocation(1).
 *      or was already allocated. If the queue was previously allocated this parameter returns 
 *      the reference count.
 *
 *  @retval
 *      Success -   Queue Handle. Used as an input parameter for operation on this queue.
 *
 *  @retval
 *      Failure -   <0
 */
Qmss_QueueHnd Qmss_queueOpenInRange (uint32_t startQueNum, uint32_t endQueNum, uint8_t *isAllocated)
{
    void               *key;
    uint32_t           index;
    int                foundQueue;
    int32_t            qType;
    int32_t            queTypeEnd;
    Qmss_QueueHnd      retVal = (Qmss_QueueHnd)(-1);

    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;

    if ((qType = Qmss_internalFindQueueType(startQueNum)) < 0)
    {
        /* startQueNum does not map to valid queue type */
        return QMSS_INVALID_PARAM;
    }

    queTypeEnd = qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(startQueNum)][qType].startIndex + 
                 qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(startQueNum)][qType].maxNum;
    if (endQueNum > queTypeEnd)
    {
        /* Queue range crosses queue type boundary */
        return QMSS_INVALID_PARAM;
    }
 
    /* Begin Critical Section before accessing shared resources. */
    key = Qmss_osalCsEnter ();

    /* Invalidate qmssQueueFree Object */
    Qmss_osalBeginMemAccess ((void *) &qmssQueueFree[startQueNum], endQueNum - startQueNum);

    *isAllocated = 0;
    foundQueue = 0;

    for (index = startQueNum; index < endQueNum; index++)
    {
        /* If the queue is alread open in this context, skip it */
        if (! qmssQueueFree[index])
        {
            if (qmssLObj.p.qmRmServiceHandle)
            {
                /* Check with RM and see if somebody else has it */
                if (Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_USE,
                                   qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(index)][qType].rmQueue, (int32_t *)&index, 1, 0, isAllocated))
                {
                    if (*isAllocated != 1)
                    {
                        /* Don't like this queue, some other context got it, give it back */
                        if (! Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_FREE,
                                             qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(index)][qType].rmQueue, (int32_t *)&index, 1, 0, isAllocated))
                        {
                            retVal = QMSS_RM_INTERNAL_ERROR;
                            goto exitCs;
                        }
                    }
                    else
                    {
                        /* Nobody else has this queue, take it */
                        foundQueue = 1;
                    }
                }
            }
            else
            {
                /* Non-RM case */
                *isAllocated = 1;
                foundQueue = 1;
            }
        }

        if (foundQueue)
        {
            qmssQueueFree[index]++;
            /* Writeback qmssQueueFree Object */
            Qmss_osalEndMemAccess ((void *) &qmssQueueFree[index], QMSS_MAX_CACHE_ALIGN);

            /* Balance the groups */
            Qmss_internalUpdateGroupControl (QMSS_QUEUE_GROUP(index), -1);

            retVal = (Qmss_QueueHnd) index;

            break;
        }
    }  

exitCs:
    /* End Critical Section */   
    Qmss_osalCsExit (key);

    return retVal;
}

/**
 *  @b Description
 *  @n  
 *      This function opens a contiguous block of queues of type queType aligned
 *      to align.  This API may fail even if sufficient queues are available
 *      due to fragmentation; thus it should be called soon after boot.
 *
 *  @param[out] queueHnds
 *      List of queue handles.  This must be pre-allocated to at least
 *      numQ entries.
 *
 *  @param[in]  queGroup
 *      Specifies the queue manager group to find a queue.  This must be specified
 *      in SPLIT mode.  It may be specified (not QMSS_PARAM_NOT_SPECIFIED) in JOINT
 *      mode to allocate a queue in the same group as another queue.
 *
 *  @param[in]  queType
 *      Specifies the type of queues that should be opened. 
 * 
 *  @param[in]  numQ
 *      Specifies the number of queues that should be opened
 *      
 *  @param[in]  align
 *      Queue number of first queue shall be such that queue number % align == 0.
 *
 *  @retval
 *      Success -   Queue Handle of first queue
 *  @retval
 *      Failure -   QMSS_RESOURCE_ALLOCATE_USE_DENIED 
 *  @retval
 *      Failure -   <0
 */
Qmss_QueueHnd Qmss_queueBlockOpenInGroup (Qmss_QueueHnd *queueHnds, int32_t queGroup, Qmss_QueueType queType, int32_t numQ, int32_t align)
{
    int queueNum, baseQueue, firstQ, lastQ;
    uint8_t isAllocated;
    void *key;
    Qmss_QueueHnd errorCode;
    int errorFound;
    uint32_t openQGroup;
    int qGroupIsSpecified;
    Qmss_QueueHnd retVal = QMSS_SOK;

    if (! queueHnds)
    {
        return QMSS_INVALID_PARAM;
    }

    if (queType == QMSS_PARAM_NOT_SPECIFIED)
    {
        return QMSS_INVALID_PARAM;
    }

    if (qmssLObjIsValid == 0)
    {
        return QMSS_NOT_INITIALIZED;
    }

    /* Convert signed PARAM_NOT_SPECIFIED into two arguments.  0 is legal queue group. */
    openQGroup        = (uint32_t)queGroup;
    qGroupIsSpecified = 1;
    if (queGroup == QMSS_PARAM_NOT_SPECIFIED) 
    {
        openQGroup = 0;
        qGroupIsSpecified = 0;
    }

    /* Begin Critical Section before accessing shared resources. */
    key = Qmss_osalCsEnter ();

    if (qmssLObj.p.qmRmServiceHandle)
    {         
        queueHnds[0] = QMSS_RESOURCE_ALLOCATE_USE_DENIED;
        baseQueue = RM_RESOURCE_BASE_UNSPECIFIED;
        if (Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_USE,
                           qmssLObj.p.maxQueueNum[openQGroup][queType].rmQueue, (int32_t *)&baseQueue, numQ, align, 
                           &isAllocated))
        {
            /* Invalidate qmssQueueFree Object */
            Qmss_osalBeginMemAccess ((void *) &qmssQueueFree[baseQueue], numQ);
            for (queueNum = 0; queueNum < numQ; queueNum++)
            {
                int thisQueue = baseQueue++;
                queueHnds[queueNum] = (Qmss_QueueHnd) thisQueue;
                if (qmssQueueFree[thisQueue])
                {
                    /* book keeping in qmssQueueFree mismatches RM */
                    retVal = QMSS_RM_INTERNAL_ERROR;
                }
                qmssQueueFree[thisQueue]++;
            }
            /* writeback qmssQueueFree Object */
            Qmss_osalEndMemAccess ((void *) &qmssQueueFree[baseQueue], numQ);
            Qmss_internalUpdateGroupControl (openQGroup, -numQ);
        }
    }
    else
    {
        firstQ = qmssLObj.p.maxQueueNum[openQGroup][queType].startIndex;
        lastQ  = qmssLObj.p.maxQueueNum[openQGroup][queType].startIndex + qmssLObj.p.maxQueueNum[openQGroup][queType].maxNum;
        queueHnds[0] = QMSS_QUEUE_OPEN_ERROR;


        /* Allocate a contigous block of numQ queues aligned to align */
        for (baseQueue = ((firstQ + align - 1) / align) * align;
             baseQueue < (lastQ - numQ);
             baseQueue += align) 
        {
            for (queueNum = 0; queueNum < numQ; queueNum++) 
            {
                queueHnds[queueNum] = 
                    Qmss_internalQueueOpen (qGroupIsSpecified, openQGroup, queType, baseQueue + queueNum, &isAllocated);
                errorFound = 0;
                if (queueHnds[queueNum] < 0)
                {
                    /* Open failed close the rest of the queues (except error queue) */
                    errorCode = queueHnds[queueNum];
                    queueNum--;
                    errorFound = 1;
                }
                if (isAllocated != 1) 
                {
                    /* Somebody else got the queue. Close what we got, and try next range */
                    errorCode = QMSS_QUEUE_OPEN_ERROR;
                    errorFound = 1;
                }
                if (errorFound) 
                {
                    /* Close any nonerrored queues */
                    for (; queueNum >= 0; queueNum--)
                    {
                        Qmss_internalQueueClose (queueHnds[queueNum]);
                        queueHnds[queueNum] = errorCode;
                    }
                    break;
                }
            }
            if (queueNum == numQ) 
            {
                break; /* successful */
            }
        }

    }

    Qmss_osalCsExit (key);

    if (retVal == QMSS_SOK)
    {
        return retVal = queueHnds[0];
    }

    return retVal;
} /* Qmss_queueBlockOpen */

/**
 *  @b Description
 *  @n  
 *      This function opens a contiguous block of queues of type queType aligned
 *      to align in queue group 0.  This API may fail even if sufficient queues are available
 *      due to fragmentation; thus it should be called soon after boot.
 *
 *  @param[out] queueHnds
 *      List of queue handles.  This must be pre-allocated to at least
 *      numQ entries.
 *
 *  @param[in]  queType
 *      Specifies the type of queues that should be opened. 
 * 
 *  @param[in]  numQ
 *      Specifies the number of queues that should be opened
 *      
 *  @param[in]  align
 *      Queue number of first queue shall be such that queue number % align == 0.
 *
 *  @retval
 *      Success -   Queue Handle of first queue
 *  @retval
 *      Failure -   QMSS_RESOURCE_ALLOCATE_USE_DENIED 
 *  @retval
 *      Failure -   <0
 */
Qmss_QueueHnd Qmss_queueBlockOpen (Qmss_QueueHnd *queueHnds, Qmss_QueueType queType, int32_t numQ, int32_t align)
{
    return Qmss_queueBlockOpenInGroup (queueHnds, QMSS_PARAM_NOT_SPECIFIED, queType, numQ, align);
}

/**
 *  @b Description
 *  @n  
 *      This function opens the specified queue if and only if it is already open elsewhere.
 *      OpenUse means "open to use queue already allocated by someone else".
 *
 *  @param[in]  QID
 *      Specifies the queue number that should be opened.
 *      
 *  @param[out]  useCount
 *      Reference count, which must be >= 2.
 *
 *  @retval
 *      Success -   Queue Handle 
 *  @retval
 *      Failure -   QMSS_RESOURCE_ALLOCATE_USE_DENIED 
 *      Failure -   QMSS_QUEUE_NOT_ALREADY_OPEN 
 *  @retval
 *      Failure -   <0
 */
Qmss_QueueHnd Qmss_queueOpenUse (uint32_t QID, uint8_t *useCount)
{
    Qmss_QueueHnd retVal;
    void          *key;
    int           tryOpening;

    if (qmssLObjIsValid == 0)
    {
        return QMSS_NOT_INITIALIZED;
    }

    /* Begin Critical Section before accessing shared resources. */
    key = Qmss_osalCsEnter ();

    /* Invalidate qmssQueueFree before using */
    Qmss_osalBeginMemAccess ((void *) &qmssQueueFree[QID], QMSS_MAX_CACHE_ALIGN);

    if (qmssQueueFree[QID] == 0)
    {
        if (qmssLObj.p.qmRmServiceHandle)
        {
            /* implementation of RM requires trial and error */
            tryOpening = 1;
        }
        else
        {
            /* No need to try the queue -- its not already open */
            tryOpening = 0;
            retVal     = QMSS_QUEUE_NOT_ALREADY_OPEN;
        }
    } else {
        /* Queue is already open in this context */
        tryOpening = 1;
    }

    if (tryOpening)
    {
        retVal = Qmss_internalQueueOpen (0, 0, (Qmss_QueueType)QMSS_PARAM_NOT_SPECIFIED, QID, useCount);
        if ((retVal >= QMSS_SOK) && (*useCount == 1) && (qmssQueueFree[QID] == 1))
        {
            /* The qmssQueueFree[QID] isn't needed if rm returns use count instead of owner count */
            Qmss_internalQueueClose (retVal);
            *useCount = 0;
            retVal    = QMSS_QUEUE_NOT_ALREADY_OPEN;
        }
        else
        {
            if (*useCount == 1) 
            {
                *useCount = 2; /* spoof 2 for RM */
            }
        }
    }

    /* End Critical Section */   
    Qmss_osalCsExit (key);

    return retVal;
} /* Qmss_queueOpenUse */

/* Internal function to close a queue, must be inside a critical section */
Qmss_Result Qmss_internalQueueClose (Qmss_QueueHnd hnd)
{
    uint8_t     refCount;
    Qmss_Result retVal;

    if (qmssLObj.p.qmRmServiceHandle)
    {       
        int32_t qType;
        
        /* Get the queue type */
        qType = Qmss_internalFindQueueType(hnd);
        
        if (Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_FREE, 
                           qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(hnd)][qType].rmQueue, &hnd, 1, 0, &refCount))
        {
            retVal = refCount;
        }
        else
        {
            retVal = QMSS_RESOURCE_FREE_DENIED;
            goto exit; /* close failed don't do internal book keeping */
        }
    }

    /* Invalidate qmssQueueFree Object */
    Qmss_osalBeginMemAccess ((void *) &qmssQueueFree[hnd], QMSS_MAX_CACHE_ALIGN);
     
    if (qmssQueueFree[hnd] == 0)
    {
        if (qmssLObj.p.qmRmServiceHandle)
        {       
            /* Out of sync with RM */
            retVal = QMSS_RM_INTERNAL_ERROR;
        }
        else
        {
            /* No RM, close failed */
            retVal = QMSS_INVALID_PARAM;
        }
    }
    else
    {
        qmssQueueFree[hnd]--;

        if (! qmssQueueFree[hnd]) 
        {
            /* When last reference is freed, increase free queues for group */
            Qmss_internalUpdateGroupControl (QMSS_QUEUE_GROUP(hnd), +1);
        }

        /* Writeback qmssQueueFree Object */
        Qmss_osalEndMemAccess ((void *) &qmssQueueFree[hnd], QMSS_MAX_CACHE_ALIGN);

        if (! qmssLObj.p.qmRmServiceHandle)
        {       
            /* When using RM, use the RM's owner count */
            retVal = qmssQueueFree[hnd];
        }
    }

exit:
    return retVal;
} /* Qmss_internalQueueClose */

/**
 *  @b Description
 *  @n  
 *      This function closes a queue. 
 *      The queue reference count is decremented. The queue is freed only if the 
 *      reference count is zero. 
 *
 *  @param[in]  hnd
 *      Queue handle.
 * 
 *  @pre  
 *      Qmss_queueOpen function should be called before calling this function.
 *
 *  @post  
 *      Queue is freed if the reference count is zero and available for reallocation.
 *
 *  @retval
 *      Success -   Current reference count.
 *  @retval
 *      Failure -   QMSS_INVALID_PARAM
 */
Qmss_Result Qmss_queueClose (Qmss_QueueHnd hnd)
{
    void        *key;
    Qmss_Result retVal;

    /* Begin Critical Section before accessing shared resources. */
    key = Qmss_osalCsEnter ();

    retVal = Qmss_internalQueueClose (hnd);

    /* End Critical Section */   
    Qmss_osalCsExit (key);

    return retVal;
} /* Qmss_queueClose */

/**
 *  @b Description
 *  @n  
 *      This function returns the threshold at which the queue threshold pin is asserted
 * 
 *  @param[in]  hnd
 *      Queue handle.
 *
 *  @pre  
 *      Qmss_queueOpen function should be called before calling this function.
 *
 *  @retval
 *      Queue threshold
 */
uint32_t Qmss_getQueueThreshold (Qmss_QueueHnd hnd)
{
    /* No RM check since read only */

    uint32_t qGroup  = QMSS_QUEUE_GROUP(hnd);
    uint32_t qNumber = QMSS_QUEUE_NUMBER(hnd);

    return (uint32_t) CSL_FEXT (qmssLObj.p.groupRegs[qGroup].qmQueStatReg->QUEUE_STATUS_CONFIG_GROUP[qNumber].QUEUE_STATUS_CONFIG_REG_D,
                    QM_QUEUE_STATUS_CONFIG_QUEUE_STATUS_CONFIG_REG_D_THRESHOLD); 
}

/**
 *  @b Description
 *  @n  
 *      This function sets the threshold at which the queue threshold pin is asserted and 
 *      threshold bit in threshold status register is set. 
 *      "hilo" indicates whether the number of items in the queue should be greater than/equal to 
 *      OR less than the confgiured threshold value before the queue ecnt status bit is asserted and
 *      threshold bit in threshold status register is set.
 *      The threshold value is 10'h3ff when it is ten or higher. It is (2^threshold-1)  for other values.
 * 
 *  @param[in]  hnd
 *      Queue handle.
 *
 *  @param[in]  hilo
 *      1 - High
 *      0 - Low.
 *
 *  @param[in]  threshold
 *      Threshold value.
 *
 *  @pre  
 *      Qmss_queueOpen function should be called before calling this function.
 *
 *  @retval
 *      Qmss_Result
 */
Qmss_Result Qmss_setQueueThreshold (Qmss_QueueHnd hnd, uint16_t hilo, uint8_t threshold)
{
    uint32_t           qGroup  = QMSS_QUEUE_GROUP(hnd);
    uint32_t           qNumber = QMSS_QUEUE_NUMBER(hnd);
    uint32_t           temp = 0;
    
    if (threshold > 10)
        threshold = 10;
    CSL_FINS (temp, QM_QUEUE_STATUS_CONFIG_QUEUE_STATUS_CONFIG_REG_D_THRESHOLD_HILO, hilo);
    CSL_FINS (temp, QM_QUEUE_STATUS_CONFIG_QUEUE_STATUS_CONFIG_REG_D_THRESHOLD, threshold);
    
    qmssLObj.p.groupRegs[qGroup].qmQueStatReg->QUEUE_STATUS_CONFIG_GROUP[qNumber].QUEUE_STATUS_CONFIG_REG_D = temp; 
    return QMSS_SOK;
}

/**
 *  @b Description
 *  @n  
 *      API to return a group of 4 starvation count available in the group
 * 
 *      Note: if handle is unaligned to 4 queues, it will be aligned.  Thus
 *      if hnd%4 = 2, the requested result will be in pStarvationCounts[2],
 *      not pStarvationCounts[0].
 *
 *  @param[in]  hnd
 *      Queue handle.
 *      Number of starvation counts from the group
 *      Array of starvation counts. Max 4
 *
 *  @param[in] numCounts
 *      Maximum number of elements in pStarvationCounts
 *
 *  @param[out] pStarvationCounts
 *      Array of starvation counts (return value)
 *
 *  @pre
 *      Qmss_queueOpen function should be called before calling this function.
 *
 *  @retval
 *      None
 */
void Qmss_getStarvationCounts (Qmss_QueueHnd hnd,
                               uint32_t      numCounts,
                               uint32_t*     pStarvationCounts)
{
    uint32_t starvationCount;
    uint32_t qGroup = QMSS_QUEUE_GROUP(hnd);
    uint32_t queNum = QMSS_QUEUE_NUMBER(hnd);
    uint32_t regIndex;

    /* No RM check since read only */
    queNum -= QMSS_QUEUE_NUMBER(qmssLObj.p.maxQueueNum[qGroup][Qmss_QueueType_STARVATION_COUNTER_QUEUE].startIndex);
    regIndex = queNum / 4; /* this will align an queue number to 4 queues */

    starvationCount = 
      qmssLObj.p.groupRegs[qGroup].qmConfigReg->FREE_DESCRIPTOR_STARVE_COUNT_REG[regIndex];
    if (numCounts >= 1)
    {
        pStarvationCounts[0] = (uint8_t)CSL_FEXTR (starvationCount, 7, 0);
    }
    if (numCounts >= 2)
    {
        pStarvationCounts[1] = (uint8_t)CSL_FEXTR (starvationCount, 15, 8);
    }
    if (numCounts >= 3)
    {
        pStarvationCounts[2] = (uint8_t)CSL_FEXTR (starvationCount, 23, 16);
    }
    if (numCounts >= 4)
    {
        pStarvationCounts[3] = (uint8_t)CSL_FEXTR (starvationCount, 31, 24);
    }
}
/**
 *  @b Description
 *  @n  
 *      The starvation count is incremented every time the Free Descriptor/Buffer 
 *      queue is read when empty. This function returns the starvation count of queue.
 * 
 *  @param[in]  hnd
 *      Queue handle.
 *
 *  @pre  
 *      Qmss_queueOpen function should be called before calling this function.
 *
 *  @retval
 *      8 bit Starvation count
 */
uint32_t Qmss_getStarvationCount (Qmss_QueueHnd hnd)
{
    uint32_t queNum = QMSS_QUEUE_NUMBER(hnd);
    uint32_t regIndex;
    uint32_t starvationCount[4];

    /* No RM check since read only */
    Qmss_getStarvationCounts(hnd,sizeof(starvationCount)/sizeof(uint32_t),starvationCount);
    regIndex = queNum % 4;
    return (uint32_t)(starvationCount[regIndex]);
}

/**
 *  @b Description
 *  @n
 *      This function returns the threshold status of the queue. There are 2 conditions under which the 
 *      threshold bit is set
 *
 *      The threshold bit is set for a queue if the number of element in a queue is above or below 
 *      a certain threshold number of items configured using Qmss_setQueueThreshold() API. 
 *
 *      The threshold bit is set for a queue if there is atleast 1 element on the queue when the threshold 
 *      is not set Qmss_setQueueThreshold() API
 *
 *  @param[in]  hnd
 *      Queue handle.
 *
 *  @pre  
 *      Qmss_queueOpen function should be called before calling this function.
 *
 *  @retval
 *      0 - Configured threshold is not met
 *  @retval
 *      1 - Configured threshold is met
 */
uint16_t Qmss_getQueueThresholdStatus (Qmss_QueueHnd hnd)
{
    uint32_t qGroup  = QMSS_QUEUE_GROUP(hnd);
    uint32_t qNumber = QMSS_QUEUE_NUMBER(hnd);
    uint32_t regIndex, bitField;

    /* No RM check since read only */

    regIndex = qNumber / 32;
    bitField = qNumber % 32;

    return (uint16_t) CSL_FEXTR (qmssLObj.p.groupRegs[qGroup].qmStatusRAM->QUEUE_THRESHOLD_STATUS_REG[regIndex], bitField, bitField);
}

/** 
 *  @b Description
 *  @n  
 *      Given the queue handle the function returns the queue manager and queue number within
 *      the queue manager corresponding to the handle
 *
 *  @param[in]  hnd      
 *      Queue handle
 *
 *  @retval
 *      Queue Manager and Queue Number within the Queue Manager
 */
Qmss_Queue Qmss_getQueueNumber (Qmss_QueueHnd hnd)
{
    Qmss_Queue              queue;

    /* No RM check since read only */

    queue.qMgr = CSL_FEXTR (hnd, 13, 12);
    queue.qNum = CSL_FEXTR (hnd, 11, 0);
    return (queue);
}

/** 
 *  @b Description
 *  @n  
 *      Given the queue handle the function returns the queue manager 
 *      group containing the queue.  This can be used to allocate other
 *      queues in the same group as a particular handle.  This is required
 *      if it is necessary to use @ref Qmss_queueDivert on the pair of
 *      queues.
 *
 *  @param[in]  hnd      
 *      Queue handle
 *
 *  @retval
 *      Queue group
 */
uint32_t Qmss_getQueueGroup (Qmss_QueueHnd hnd)
{
    /* No RM check since read only */
    return (QMSS_QUEUE_GROUP(hnd));
}

/** 
 *  @b Description
 *  @n  
 *      Given the queue manager and queue number within the queue manager 
 *      the function returns the corresponding queue handle
 *
 *  @param[in]  queue      
 *      Queue Manager number and Queue Number within the Queue Manager
 *
 *  @retval
 *      Queue handle corresponding to the queue manager and queue number
 */
Qmss_QueueHnd Qmss_getQueueHandle (Qmss_Queue queue)
{
    Qmss_QueueHnd           hnd = 0;

    /* No RM check since read only */

    CSL_FINSR (hnd, 13, 12, queue.qMgr);
    CSL_FINSR (hnd, 11, 0, queue.qNum);
    return (hnd);
}

/** 
 *  @b Description
 *  @n  
 *      Downloads the PDSP firmware to PDSP.
 *
 *  @param[in]  pdspId      
 *      PDSP number where the firmware should be downloaded
 *
 *  @param[in]  image      
 *      Pointer to the firmware image
 *
 *  @param[in]  size      
 *      Size of firmware image in bytes
 *
 *  @retval
 *      Success -   Firmware is downloaded to PDSP
 *  @retval
 *      Failure -   QMSS_NOT_INITIALIZED
 *  @retval
 *      Failure -   QMSS_INVALID_PARAM
 *  @retval
 *      Failure -   QMSS_RESOURCE_ALLOCATE_USE_DENIED
 */
Qmss_Result Qmss_downloadFirmware (Qmss_PdspId pdspId, void *image, uint32_t size)
{
    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;

    if ((image == NULL) || (size == 0))
        return QMSS_INVALID_PARAM;

    return (Qmss_internaldownloadFirmware (pdspId, image, size));
}

static Qmss_Result Qmss_internaldownloadFirmware (Qmss_PdspId pdspId, void *image, uint32_t size)
{
    uint32_t           i, count;
    volatile uint32_t  enable;
    uint32_t           *data = (uint32_t *) image;  

    if (qmssLObj.p.qmRmServiceHandle)
    {
        int32_t pdspIdNum = (int32_t)pdspId;
        /* Check Firmware use privileges */
        if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_USE, 
                            qmssLObj.p.rmFirmwarePdsp, &pdspIdNum, 1, 0, NULL))
        {
            return(QMSS_RESOURCE_ALLOCATE_USE_DENIED);
        }
        /* Check Firmware init privileges */
        if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_INIT,
                            qmssLObj.p.rmFirmwarePdsp, &pdspIdNum, 1, 0, NULL))
        {
            /* Use but don't init - Check firmware revision */
        }
    }

    /* Reset the PDSP */
    CSL_FINS (qmssLObj.p.regs.qmPdspCtrlReg[pdspId]->PDSP_CONTROL_REG, PDSP_PDSP_CONTROL_REG_PDSP_ENABLE, 0);

    /* Confirm PDSP has halted */
    do
    { 
        enable = CSL_FEXT (qmssLObj.p.regs.qmPdspCtrlReg[pdspId]->PDSP_CONTROL_REG, PDSP_PDSP_CONTROL_REG_PDSP_STATE);
    }while (enable);

    count = size / 4;

    /* Some firmwares need the 1 based PDSP number written to scratch + 0x18 */
    *(qmssLObj.p.regs.qmPdspCmdReg[pdspId] + 6) = pdspId + 1;

    /* Download the firmware */
    for(i = 0; i < count; i++)
        qmssLObj.p.regs.qmPdspIRamReg[pdspId][i] = data[i];

    /* Use the command register to sync the PDSP */
    *qmssLObj.p.regs.qmPdspCmdReg[pdspId] = 0xFFFFFFFF;

    /* Wait to the memory write to land */
    for(i = 0; i < 20000 && *qmssLObj.p.regs.qmPdspCmdReg[pdspId] != 0xFFFFFFFF; i++);
   
    /* Reset program counter to zero */
    CSL_FINS (qmssLObj.p.regs.qmPdspCtrlReg[pdspId]->PDSP_CONTROL_REG, PDSP_PDSP_CONTROL_REG_PCOUNTER_RST_VAL, 0);
    
    /* Set soft reset to zero to load the program counter */
    CSL_FINS (qmssLObj.p.regs.qmPdspCtrlReg[pdspId]->PDSP_CONTROL_REG, PDSP_PDSP_CONTROL_REG_SOFT_RST_N, 0);
    
    /* Enable the PDSP */
    CSL_FINS (qmssLObj.p.regs.qmPdspCtrlReg[pdspId]->PDSP_CONTROL_REG, PDSP_PDSP_CONTROL_REG_PDSP_ENABLE, 1);

    /* Wait for the command register to clear */
    while (*qmssLObj.p.regs.qmPdspCmdReg[pdspId]);

    return QMSS_SOK;
}

/**
 *  @b Description
 *  @n  
 *      This function sets the end of interrupt vector value for the specified interrupt. 
 *      This API is used by the host to indicate 
 *          - The end of interrupt processing to the accumulator. The accumulator 
 *              is then free to write to the accumulator ping/pong page released by the host.
 *          - The end of interrupt processing for buffer descriptor starvation event on 
 *              receive SOP and MOP for any of the RX DMA units in the CDMA.
 *
 *  @param[in]  intdNum
 *      Which QMSS INTD to program.  As this is a fast path API it is NOT bounds checked.
 *
 *  @param[in]  type
 *      Interrupt  to program. 
 *          Qmss_IntdInterruptType_HIGH selects the interrupts corresponding to the high priority accumulator
 *          Qmss_IntdInterruptType_LOW selects the interrupts corresponding to the low priority accumulator
 *          Qmss_IntdInterruptType_CDMA selects the interrupts corresponding to the low priority accumulator
 *
 *  @param[in]  accumCh
 *      Accumulator channel which is mapped to interrupt number via Qmss_xxxEoiVector mapping 
 *      table.  The resulting interrupt number is set. 
 *
 *      **No validation is done to check if the interrupt number is within range to reduce overhead.**
 *
 *  @post  
 *      EOI vector value is set indicating end of host interrupt processing.
 *
 *  @retval
 *      Success - QMSS_ACC_SOK
 *  @retval
 *      Failure - QMSS_NOT_INITIALIZED
 */

Qmss_Result Qmss_setEoiVectorByIntd (uint8_t intdNum, Qmss_IntdInterruptType type, uint8_t accumCh)
{ 
    CSL_Qm_intdRegs *intd;

    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;

    intd = qmssLObj.p.regs.qmQueIntdReg[intdNum];

    /* It is assumed that all intd's share the same channel to vector map.  If this ever
     * varies, Qmss_*EoiVector[] needs to be indexted by intd as well */
    if (type == Qmss_IntdInterruptType_HIGH)
        CSL_FINS (intd->EOI_REG, QM_INTD_EOI_REG_EOI_VECTOR, Qmss_highEoiVector[accumCh]);
    else if (type == Qmss_IntdInterruptType_LOW)
        CSL_FINS (intd->EOI_REG, QM_INTD_EOI_REG_EOI_VECTOR, Qmss_lowEoiVector[accumCh]);
    else 
        CSL_FINS (intd->EOI_REG, QM_INTD_EOI_REG_EOI_VECTOR, Qmss_cdmaEoiVector[accumCh]);
    return QMSS_SOK;
}

/**
 *  @b Description
 *  @n  
 *      Calls @ref Qmss_setEoiVectorByIntd for the first intd in QMSS
 *      (for backwards compatibility).
 *
 *  @retval
 *      Success - QMSS_ACC_SOK
 *  @retval
 *      Failure - QMSS_NOT_INITIALIZED
 */
Qmss_Result Qmss_setEoiVector (Qmss_IntdInterruptType type, uint8_t accumCh)
{
    return Qmss_setEoiVectorByIntd (0, type, accumCh);
} /* Qmss_setEoiVector */

/**
 *  @b Description
 *  @n  
 *      This function sets the end of interrupt vector value for the specified interrupt. 
 *      This API is used by the host to indicate 
 *              host is done processing interrupt number intNum.  This API is normally 
 *              not used.  See @ref Qmss_setEoiVectorByIntd.
 *
 *  @param[in]  intdNum
 *      Which QMSS INTD to program
 *
 *  @param[in]  intNum
 *      Interrupt number to program. 
 *
 *  @post  
 *      EOI vector value is set indicating end of host interrupt processing.
 *
 *  @retval
 *      Success - QMSS_SOK
 *  @retval
 *      Failure - QMSS_NOT_INITIALIZED
 */
Qmss_Result Qmss_setEoiIntNumByIntd (uint8_t intdNum, uint8_t intNum)
{ 
    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;

    CSL_FINS (qmssLObj.p.regs.qmQueIntdReg[intdNum]->EOI_REG, QM_INTD_EOI_REG_EOI_VECTOR, intNum);

    return QMSS_SOK;
} /* Qmss_setEoiIntNumByIntd */

/**
 *  @b Description
 *  @n Calls @ref Qmss_setEoiIntNumByIntd with intdNum = 0 (for backwards compatibility)
 *
 *  @retval
 *      Success - QMSS_SOK
 *  @retval
 *      Failure - QMSS_NOT_INITIALIZED
 */
Qmss_Result Qmss_setEoiIntNum (uint8_t intNum)
{
    return Qmss_setEoiIntNumByIntd (0, intNum);
} /* Qmss_setEoiIntNum */

/**
 *  @b Description
 *  @n  
 *      This function decrements the interrupt count. The Interrupt Count Registers contains a 
 *      count of the interrupts that have triggered and not been processed. Writing a non-zero 
 *      value to the register subtracts that value from the count. Writing a 0 clears the count. 
 *      The count saturates at 3.
 *
 *      Note: Clearing the count does not clear the interrupt. The EOI Vector must be set using Qmss_setEoiVector API.
 *  
 *  @param[in]  intdNum
 *      Which QMSS INTD to program.  As this is a fast path API it is NOT bounds checked.
 *
 *  @param[in]  interruptNum
 *      Interrupt Number that should be acknowledged. 
 *
 *  @param[in]  value
 *      Value to decrement by or zero to clear. 
 *      
 *      **No validation is done to check if the interrupt number is within range to reduce overhead.**
 *
 *  @post  
 *      Count is decremented or cleared depending on input parameter. 
 *
 *  @retval
 *      Success - QMSS_ACC_SOK
 *  @retval
 *      Failure - QMSS_NOT_INITIALIZED
 */

Qmss_Result Qmss_ackInterruptByIntd (uint8_t intdNum, uint8_t interruptNum, uint8_t value)
{
    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;
    
    qmssLObj.p.regs.qmQueIntdReg[intdNum]->INTCNT_REG[interruptNum] = value;
    return QMSS_SOK;
} /* Qmss_ackInterruptByIntd */

/**
 *  @b Description
 *  @n Calls @ref Qmss_ackInterruptByIntd with intdNum = 0 (for backwards compatibility)
 *
 *  @retval
 *      Success - QMSS_SOK
 *  @retval
 *      Failure - QMSS_NOT_INITIALIZED
 */
Qmss_Result Qmss_ackInterrupt (uint8_t interruptNum, uint8_t value)
{
    return Qmss_ackInterruptByIntd (0, interruptNum, value);
} /* Qmss_ackInterrupt */

/**
 *  @b Description
 *  @n  
 *      The function is used to get the version information of the QMSS LLD.
 *
 *  @retval
 *      Version Information.
 */
uint32_t Qmss_getVersion (void)
{
    return QMSS_LLD_VERSION_ID;
}

/**
 *  @b Description
 *  @n  
 *      The function is used to get the version string for the QMSS LLD.
 *
 *  @retval
 *      Version String.
 */
const char* Qmss_getVersionStr (void)
{
    return qmssLldVersionStr;
}

/**
@}
*/

/** 
 *  @b Description
 *  @n  
 *      The function return the source queue handle on which the descriptors configured for 
 *      the specified memory region are stored.
 *
 *      This is for *internal* LLD usage.
 *
 *  @param[in]  memRegion      
 *      Memory region number
 *
 *  @retval
 *      Success -   Queue Handle
 *  @retval
 *      Failure -   QMSS_NOT_INITIALIZED
 */
int32_t Qmss_getMemRegQueueHandle(uint32_t memRegion)
{
    /* No RM check since internal LLD usage */
  
    if (qmssLObjIsValid == 0)
        return QMSS_NOT_INITIALIZED;
    else
    {
        /* Invalidate Global Object */
        Qmss_osalBeginMemAccess ((void *) &qmssGObj.obj, sizeof (Qmss_GlobalObj_Unpadded));
        return (qmssGObj.obj.descQueue[memRegion]);
    }
}

/** 
 *  @b Description
 *  @n  
 *      The function returns size of the descriptors in the specified memory region.
 *
 *      This is for *internal* LLD usage.
 *
 *  @param[in]  memRegion      
 *      Memory region number.
 *  @param[in]  qGroup      
 *      Queue manager group number
 *
 *  @retval
 *      Descriptor size
 */
uint32_t Qmss_getMemRegDescSize (uint32_t memRegion, uint32_t qGroup)
{
    uint32_t  descVal;

    /* No RM check since internal LLD usage */
        
    descVal = CSL_FEXT (qmssLObj.p.groupRegs[qGroup].qmDescReg->MEMORY_REGION_BASE_ADDRESS_GROUP[memRegion].MEMORY_REGION_DESCRIPTOR_SETUP_REG, 
            QM_DESCRIPTOR_REGION_CONFIG_MEMORY_REGION_DESCRIPTOR_SETUP_REG_DESC_SIZE);
    return ((descVal + 1) * 16);
}

/**
 *  @b Description
 *  @n  
 *      Internal queue open function; is called to allocate a queue when critical section is provided 
 *      by the caller. 
 *
 *      A queue can be opened in two ways:
 *          1) If "queNum" is set to QMSS_PARAM_NOT_SPECIFIED, then a new
 *          available queue of type "queType" is allocated.
 *          2) If "queNum" is a valid queue number i.e., >= 0, 
 *               then a queue is allocated if free 
 *               else
 *               a handle to a previously opened queue is returned.
 *               In this case "queType" parameter is not used.
 *
 *  @param[in]  qGroupIsSpecified
 *      0: Ignore qGroup in joint mode and allocate per balancing scheme
 *      1: Use gGroup in joint mode
 *
 *  @param[in]  qGroup
 *      Specifies queue group to open from (only used in SPLIT mode 
 *
 *  @param[in]  queType
 *      Specifies the type of queue that should be opened. 
 * 
 *  @param[in]  queNum
 *      Specifies the queue number that should be opened.
 *      
 *  @param[out]  isAllocated
 *      Indicates whether the requested queue is a new queue allocation(1).
 *      or was already allocated. If the queue was previously allocated this parameter returns 
 *      the reference count.
 *
 *  *  @pre  
 *      Appropriate Critical section protection should be provided by caller.
 *
 *  @retval
 *      Success -   Queue Handle. Used as an input parameter for operation on this queue.
 *  @retval
 *      Failure -   QMSS_RESOURCE_ALLOCATE_USE_DENIED
 *  @retval
 *      Failure -   <0
 */
static Qmss_QueueHnd Qmss_internalQueueOpen (int qGroupIsSpecified, uint32_t qGroup, Qmss_QueueType queType, int queNum, uint8_t *isAllocated)
{
    int32_t       index;
    uint32_t      startQueue, endQueue;
    uint32_t      firstQGroupIdx, qGroupIdx, lastQGroupIdx;
    uint32_t      firstQType, regQType, lastQType;
    uint32_t      qGroupSearch[QMSS_MAX_QMGR_GROUPS];
    uint32_t      nGroups = qmssLObj.p.maxQueMgrGroups;
    int           foundQ;
   
    *isAllocated = 0;

    /* If queue is already specified just open it */
    if (queNum >= 0)
    {
        if (queType == QMSS_PARAM_NOT_SPECIFIED)
        {
            if ((queType = (Qmss_QueueType) Qmss_internalFindQueueType(queNum)) < 0)
            {
                /* queNum does not map to valid queue type */
                return QMSS_INVALID_PARAM;
            }
        }
        
        if (qmssLObj.p.qmRmServiceHandle)
        {            
            /* Use QMSS_QUEUE_GROUP to find que group because the qGroup may not always be specified */
            if (!Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_USE, 
                                qmssLObj.p.maxQueueNum[QMSS_QUEUE_GROUP(queNum)][queType].rmQueue, &queNum, 1, 0, isAllocated))
            {
                return QMSS_RESOURCE_ALLOCATE_USE_DENIED;
            }
        }

        /* Invalidate qmssQueueFree Object */
        Qmss_osalBeginMemAccess ((void *) &qmssQueueFree[queNum], QMSS_MAX_CACHE_ALIGN);

        qmssQueueFree[queNum]++;

        /* With RM, the isAllocated is the "owner count" */
        if (! qmssLObj.p.qmRmServiceHandle)
        {
            *isAllocated = qmssQueueFree[queNum];
        }

        /* Internal book keeping always uses internal ref count */
        if (qmssQueueFree[queNum] == 1)
        {
            /* If new queue is opened then count it against its group */
            Qmss_internalUpdateGroupControl (QMSS_QUEUE_GROUP(queNum), -1);
        }

        /* Writeback qmssQueueFree Object */
        Qmss_osalEndMemAccess ((void *) &qmssQueueFree[queNum], QMSS_MAX_CACHE_ALIGN);

        return (Qmss_QueueHnd) queNum;
    }

    if (queType == QMSS_PARAM_NOT_SPECIFIED)
    {
        /* Queue type must be specified when queNum is not specified */
        return QMSS_INVALID_PARAM;
    }


    /* Get the group list to allocate from */
    if ((qmssLObj.mode == Qmss_Mode_SPLIT) || (qGroupIsSpecified)) 
    {
        /* Search only specified group */
        firstQGroupIdx  = 0;
        lastQGroupIdx   = 0;
        qGroupSearch[0] = qGroup;
    }
    else 
    {
        /* Search all queue groups using requested schedule */
        firstQGroupIdx = 0;
        lastQGroupIdx  = nGroups - 1;
        /* Invalidate the qmssGroupControl structure */
        Qmss_osalBeginMemAccess ((void *) &qmssGroupControl, sizeof(qmssGroupControl));
        if (qmssLObj.mode == Qmss_Mode_JOINT_ROUNDROBIN)
        {
            uint32_t alternate = qmssGroupControl.alternate;
            uint32_t i;
            /* Allocate first from QM specified by qmssGroupFree */
            for (qGroupIdx = alternate, i = 0; i < nGroups; i++) 
            {
                qGroupSearch[qGroupIdx] = qGroupIdx;
                qGroupIdx++;
                if (qGroupIdx >= nGroups)
                {
                    qGroupIdx = 0;
                }
            }
            /* Move to next QM, regardless on allocation success */
            alternate++;
            if (alternate >= nGroups)
            {
                alternate = 0;
            }
            qmssGroupControl.alternate = alternate;
        }
        else
        {
            uint32_t i,j,temp;
            /* Find lowest used QM (with most free queues) and try to allocate there first */

            /* Start by initializing the group indicies in order */
            for (qGroupIdx = 0; qGroupIdx < nGroups; qGroupIdx ++)
            {
               qGroupSearch[qGroupIdx] = qGroupIdx;
            }

            /* Sort the indicies by number of free queues.  Inefficient sort
             * used since only 1-4 groups are expected
             */
            for (i = 0; i < nGroups; i++)
            {
                for (j = i + 1; j < nGroups; j++)
                {
                    if (qmssGroupControl.groupFreeQueues[qGroupSearch[i]] <
                        qmssGroupControl.groupFreeQueues[qGroupSearch[j]])
                    {
                        temp = qGroupSearch[i];
                        qGroupSearch[i] = qGroupSearch[j];
                        qGroupSearch[j] = temp;
                    }
                }
            }
        }
        /* Writeback qmssGroupControl structure */
        Qmss_osalEndMemAccess ((void *) &qmssGroupControl, sizeof(qmssGroupControl));
    }
    /* Now qGroupSearch contains the list of queue groups to try to allocate
     * from in the priority order.
     */

    /* Get the type range to allocate from (to support multiple areas of general purpose queues) */
    if (queType == Qmss_QueueType_GENERAL_PURPOSE_QUEUE) 
    {
         /* Search all the general purpose queue ranges */
         firstQType = (uint32_t)Qmss_QueueType_GENERAL_PURPOSE_QUEUE;
         lastQType  = QMSS_MAX_QUEUE_TYPES - 1;
    }
    else
    {
         /* Search only specified queue type */
         firstQType = (uint32_t)queType;
         lastQType  = (uint32_t)queType;
    }

    /* Try each permitted queue group */
    for (qGroupIdx = firstQGroupIdx; qGroupIdx <= lastQGroupIdx; qGroupIdx++)
    {
        qGroup = qGroupSearch[qGroupIdx];

        /* Try each permitted queue type (to cover more than one area of gen purpose queues */
        for (regQType = firstQType; regQType <= lastQType; regQType++)
        {
            foundQ = 0;
            
            if (qmssLObj.p.qmRmServiceHandle)
            {           
                index = RM_RESOURCE_BASE_UNSPECIFIED;
                if (Qmss_rmService((Rm_ServiceHandle *)qmssLObj.p.qmRmServiceHandle, Rm_service_RESOURCE_ALLOCATE_INIT, 
                                   qmssLObj.p.maxQueueNum[qGroup][regQType].rmQueue, (int32_t *)&index, 1, 0, isAllocated))
                {
                    foundQ = 1;

                    /* Invalidate qmssQueueFree Object */
                    Qmss_osalBeginMemAccess ((void *) &qmssQueueFree[index], QMSS_MAX_CACHE_ALIGN);
                    if (qmssQueueFree[index] != 0)
                    {
                        return QMSS_RM_INTERNAL_ERROR;
                    }
                    qmssQueueFree[index]++;
                    /* Writeback qmssQueueFree Object */
                    Qmss_osalEndMemAccess ((void *) &qmssQueueFree[index], QMSS_MAX_CACHE_ALIGN);
                }
            }
            else
            {
                /* Get the queue range to allocate from */
                startQueue = qmssLObj.p.maxQueueNum[qGroup][regQType].startIndex;
                endQueue = qmssLObj.p.maxQueueNum[qGroup][regQType].startIndex + qmssLObj.p.maxQueueNum[qGroup][regQType].maxNum;

                /* Invalidate qmssQueueFree Object */
                Qmss_osalBeginMemAccess ((void *) &qmssQueueFree[startQueue], qmssLObj.p.maxQueueNum[qGroup][regQType].maxNum);

                for (index = startQueue; index < endQueue; index++)
                {
                    if (qmssQueueFree[index] == 0)
                    {
                        qmssQueueFree[index]++;

                        /* Writeback qmssQueueFree Object */
                        Qmss_osalEndMemAccess ((void *) &qmssQueueFree[index], QMSS_MAX_CACHE_ALIGN);

                        *isAllocated = qmssQueueFree[index];
                        foundQ = 1;
                        break;
                    }
                }
            }

            if (foundQ)
            {
                /* If new queue is opened then count it against its group */
                Qmss_internalUpdateGroupControl (QMSS_QUEUE_GROUP(index), -1);
                return (Qmss_QueueHnd) index;
            }
        }
    }

    return (Qmss_QueueHnd) -1;
}

