/**
 *   @file  fault_mgmt.c
 *
 *   @brief   
 *      The file implements the Fault Management library.
 *
 *  \par
 *  NOTE:
 *      (C) Copyright 2012 Texas Instruments, Inc.
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

 
/**************************************************************************
 *************************** Include Files ********************************
 **************************************************************************/

/* Standard Include Files. */
#include <string.h>

/* BIOS/XDC Include Files. */
#include <xdc/std.h>
#include <xdc/runtime/System.h>
#include <ti/sysbios/family/c64p/Exception.h>

/* CSL Include Files. */
#include <ti/csl/csl_chip.h>
#include <ti/csl/csl_cacheAux.h>
#include <ti/csl/csl_ipcAux.h>

/* Fault Management Include File */
#include <ti/instrumentation/fault_mgmt/fault_mgmt.h>

/* Fault Management OSAL File. */
#include <fault_mgmt_osal.h>

#define IPCGR_HOST_INDEX (8)

/**************************************************************************
 ************************** Global Variables ******************************
 **************************************************************************/

/** @addtogroup FAULT_MGMT_DATA_STRUCTURE
 @{ */

/**
 * @brief 
 *  This is the memory used by the fault management library to store any data that will be 
 *  sent to the fault handling module.
 */
#pragma DATA_ALIGN   (fault_mgmt_data, CACHE_L2_LINESIZE)
#pragma DATA_SECTION (fault_mgmt_data, ".note");
uint8_t     fault_mgmt_data[FAULT_MGMT_DATA_SIZE];

/** Note section name length in bytes (chars). */
#define NOTE_NAME_LEN 32

/** 
 * @brief Structure specifying the ELF program header Note section header.
 *           ELF Note Section:
 *
 *           byte      +0        +1       +2        +3
 *                  +--------+--------+---------+--------+
 *           namesz | # bytes in name field excluding pad|    
 *                  +--------+--------+---------+--------+
 *           descsz | # bytes in desc field excluding pad|
 *                  +--------+--------+---------+--------+
 *           type   |      Not Used (should be zero)     |
 *                  +--------+--------+---------+--------+
 *           name   | char 0 | char 1 | char 2  | char 3 |
 *                  +--------+--------+---------+--------+
 *                  | char 4 |  ...   | char n  |  pad   |      
 *                  +--------+--------+---------+--------+
 *           desc   |              word 0                |   
 *                  +--------+--------+---------+--------+
 *                  |              word 1                |
 *                  +--------+--------+---------+--------+
 */
typedef struct
{
  uint32_t namesz;
  uint32_t descsz;
  uint32_t type;
} elfNoteHeader;

/**
@}
*/

/**********************************************************************
 ********************* Fault Management Functions *********************
 **********************************************************************/

/** @addtogroup FAULT_MGMT_FUNCTION
 @{ */

/**
 *  @b Description
 *  @n  
 *      The function is used to get the register status.
 *
 *  @retval
 *      Not Applicable    
 */
void Fault_Mgmt_getLastRegStatus(void)
{
  Exception_Status status;
  elfNoteHeader noteHeader;
  char elfNoteName[NOTE_NAME_LEN];
  uint32_t *faultDataPtr = (uint32_t *)&fault_mgmt_data[0];
  uint16_t numBytesInNoteSection = 0;

  /* Initialize the fault data region to all zeros */
  memset ((void *) &fault_mgmt_data[0], 0x0, FAULT_MGMT_DATA_SIZE);

  /* Initialize number of bytes in Note Header to size of the header structure */
  numBytesInNoteSection = sizeof (elfNoteHeader);

  /* Populate Note section name using DNUM */
  System_sprintf(elfNoteName, "Core%d", DNUM);

  /* Populate Note Section header */
  noteHeader.namesz = strlen(elfNoteName) + 1; /* Add one for null character */
  /* Calculate size of desc section.  Subtract out size of pointer to Exception_Context structure
    * since that structure is added in entirety.  Add an additional uint32_t for PC value that is added. 
    * The PC is not tracked by the Exception Status or Context*/
  noteHeader.descsz = (sizeof(ti_sysbios_family_c64p_Exception_Status) -
                                   sizeof(ti_sysbios_family_c64p_Exception_Context *)) + 
                                  sizeof (ti_sysbios_family_c64p_Exception_Context) +
                                  sizeof (uint32_t); 
  noteHeader.type = 0;  /* No specific type */

  /* Format the ELF note section for the exception handler.   Increment by 4 bytes to 
    * conform to ELF header note section requirements.  Increment of 4 bytes is handled by
    * creating faultDataPtr as uint32_t. */
  *faultDataPtr++ = noteHeader.namesz;
  *faultDataPtr++ = noteHeader.descsz;
  *faultDataPtr++ = noteHeader.type;

  /* Copy the Note section name to the fault management data memory */
  memcpy ((void *) faultDataPtr, (void *)&elfNoteName[0], noteHeader.namesz);
  /* Update the data pointer and the number of bytes in the Note section, accounting for 
    * the pad at the end of the name, if needed */
  faultDataPtr += (noteHeader.namesz/sizeof(uint32_t));
  numBytesInNoteSection += noteHeader.namesz;
  if (noteHeader.namesz % sizeof(uint32_t))
  {
    faultDataPtr++;
    numBytesInNoteSection += (sizeof(uint32_t) - (noteHeader.namesz % sizeof(uint32_t)));
  }

  /* Get the register status for the core dump. */
  Exception_getLastStatus (&status);
  
  /* Copy the execution context to fault management data memory and update number
    * of bytes in the Note section. */

  /* The very first data value in the "desc" field is the PC for the crash dump.  The PC value
  * should correspond to the NMI return pointer that was stored in the NRP
  * register when the DSP exception triggered the exception handler which runs in the NMI
  * context */
  *faultDataPtr++ = status.nrp;

  /* Save the rest of the registers */
  *faultDataPtr++ = status.efr;
  *faultDataPtr++ = status.nrp;
  *faultDataPtr++ = status.ntsr;
  /* Copy IERR from exception context to the status value.  According to the exception module
    * it's tracked in two places.  However, it is only updated in one of them.  To avoid confusion on the
    * host side when the fault is reported the valid value in the exception context is copied to the invalid
    * value in the status context */
  status.ierr = (uint32_t) status.excContext->IERR;
  *faultDataPtr++ = status.ierr;
  memcpy ((void *) faultDataPtr, (void *)status.excContext, sizeof (ti_sysbios_family_c64p_Exception_Context));
  numBytesInNoteSection += noteHeader.descsz;

  /* Memory access is complete. */
  Fault_Mgmt_osalEndMemAccess(&fault_mgmt_data[0], numBytesInNoteSection);    

  return;
}

/**
 *  @b Description
 *  @n  
 *      The function is used to notify ARM the DSP core has faulted. The fault 
 *      information collected are read by the ARM from the relevant sections. 
 *
 *  @retval
 *      Not Applicable
 */
void Fault_Mgmt_notify(void)
{
  CSL_IPC_genGEMInterrupt(IPCGR_HOST_INDEX, FAULT_MGMT_HOST_IPCGR_OFFSET + DNUM);
  return;
}

/**
 *  @b Description
 *  @n  
 *      The function is used to notify DSP core that to fault by creating the
 *      NMI pulse to the GEM. The fault information collected are read by 
 *      the ARM from the relevant sections. 
 *
 *  @retval
 *      Not Applicable
 */
void Fault_Mgmt_notify_remote_core(uint32_t core_id)
{
  CSL_IPC_genNMIEvent(core_id);
  return;
}
/**
@}
*/
